from ontobio.ontol_factory import OntologyFactory
from networkx.algorithms.dag import ancestors
from ontobio.io.ontol_renderers import GraphRenderer

PLOIDY = 'PATO:0001374'

Y_SHAPED = 'PATO:0001201'

def test_remote_sparql():
    """
    reconstitution test
    """
    factory = OntologyFactory()
    # default method is sparql
    ont = factory.create('pato')
    g = ont.get_graph()
    info = g.nodes[PLOIDY]
    print(str(info))
    nodes = g.nodes()
    print(len(nodes))
    assert len(nodes) > 100
    nbrs = list(g.successors(PLOIDY))
    print("SUCC:"+str(nbrs))
    parents = list(g.predecessors(PLOIDY))
    print("PRED:"+str(parents))
    assert parents == ['PATO:0001396']
    ancs = ancestors(g, PLOIDY)
    print("ANCS:"+str(ancs))
    assert 'PATO:0000001' in ancs
    print(g)
    Q = ['.*shape.*']
    w = GraphRenderer.create('tree')

    shapes1 = ont.resolve_names(Q, is_regex=True, is_remote=False)
    print("SHAPE Q:"+str(shapes1))
    show_nodes(w, ont, shapes1)
    assert Y_SHAPED in shapes1
    
    shapes2 = ont.resolve_names(Q, is_regex=True, is_remote=True)
    print("SHAPE Q:"+str(shapes2))
    show_nodes(w, ont, shapes2)
    assert Y_SHAPED in shapes2

def show_nodes(w, ont, ids):
    for id in ids:
        print(w.render_noderef(ont, id))



from rdflib import Graph
from rdflib.plugins.sparql import prepareQuery
from ontobio.rdfgen.assoc_rdfgen import prefix_context


class RdflibSparqlWrapper:

    def run_query(self, graph: Graph, query):
        response = graph.query(prepareQuery(query, initNs=prefix_context))

        return response

    def find_involved_in_translated(self, graph: Graph, gp, term):
        # Recreate this query
        # query_pair = TriplePair((upt.molecular_function, ENABLED_BY, annoton.enabled_by),
        #                         (upt.molecular_function, PART_OF, term), connecting_entity=upt.molecular_function)
        query = """
            SELECT ?mf ?gp ?term
            WHERE {{
                ?mf rdf:type GO:0003674 .
                ?gp rdf:type {gp} .
                ?term rdf:type {term} .
                
                ?mf RO:0002333 ?gp .
                ?mf BFO:0000050 ?term
            }} 
        """.format(gp=gp, term=term)
        res = self.run_query(graph, query)
        return res

    def find_triple_by_class(self, graph: Graph, s="?any_s", p="?any_p", o="?any_o"):
        query = """
            SELECT *
            WHERE {{
                ?s rdf:type {s} .
                ?o rdf:type {o} .
                ?s {p} ?o
            }}
        """.format(s=s, p=p, o=o)
        res = self.run_query(graph, query)
        return res

    def find_acts_upstream_of_translated(self, graph, gp, causally_rel, term):
        # Recreate this query
        # query_pair = TriplePair((upt.molecular_function, ENABLED_BY, annoton.enabled_by),
        #                         (upt.molecular_function, causally_relation_uri, term),
        #                         connecting_entity=upt.molecular_function)
        query = """
            SELECT ?mf ?gp ?term
            WHERE {{
                ?mf rdf:type GO:0003674 .
                ?gp rdf:type {gp} .
                ?term rdf:type {term} .
                
                ?mf RO:0002333 ?gp .
                ?mf {causally_rel} ?term
            }}
        """.format(gp=gp, term=term, causally_rel=causally_rel)
        res = self.run_query(graph, query)
        return res

    def find_evidence_with(self, graph, annotated_source, annotated_property, annotated_target):
        # TODO: Just get all evidence properties e.g. contributors, date, reference
        query = """
            PREFIX lego: <http://geneontology.org/lego/>
        
            select *
            where {{
                ?source rdf:type {annotated_source} .
                ?target rdf:type {annotated_target} .
                                
                ?axiom owl:annotatedSource ?source .
                ?axiom owl:annotatedProperty {annotated_property} .
                ?axiom owl:annotatedTarget ?target .
                ?axiom lego:evidence ?evidence .
                
                ?evidence lego:evidence-with ?evi_with
            }}
        """.format(annotated_source=annotated_source, annotated_property=annotated_property, annotated_target=annotated_target)
        res = self.run_query(graph, query)
        return res

    def find_nested_location_chain(self, graph, primary_term_shape, *nested_terms):
        # nested_terms - Ordered list of chain elements including anchor term (e.g. root MF, primary CC)
        # Will look for chain of 'MF:root-occurs_in->term1-part_of->term2-part_of->term3-part_of->...'
        select_fields = []
        type_declarations = []
        location_edges = []
        type_declaration = "?loc_{} rdf:type {}"
        location_edge_pattern = "?loc_{} {} ?loc_{}"
        for idx, term in enumerate(nested_terms):
            term_ind_idx = idx+1
            select_fields.append("?loc_{}".format(term_ind_idx))
            type_declarations.append(type_declaration.format(term_ind_idx, term))
            edge_relation = "BFO:0000050"  # part_of
            # Here we use aspect to determine first edge_relation
            # This aspect param is a lazy hack because I haven't figured out how to do this directly in SPARQL
            if term_ind_idx == 1:
                if primary_term_shape in ["MolecularFunction", "BiologicalProcess"]:
                    edge_relation = "BFO:0000066"  # occurs_in
                elif primary_term_shape == "ProteinContainingComplex":
                    edge_relation = "RO:0001025"  # located_in
            if term_ind_idx < len(nested_terms):
                location_edges.append(location_edge_pattern.format(term_ind_idx, edge_relation, term_ind_idx+1))
        query = """
            SELECT {select_fields}
            WHERE {{
                {type_declarations} .

                {location_edges}
            }}
        """.format(select_fields=" ".join(select_fields), type_declarations=" .\n".join(type_declarations),
                   location_edges=" .\n".join(location_edges))
        # print(query)
        res = self.run_query(graph, query)
        return res



from ontobio.ecomap import EcoMap
from abc import ABC, abstractmethod
import yaml


class FilterRule(ABC):

    def __init__(self):
        self.id = None
        self.unwanted_evidence_codes = []
        self.unwanted_evi_code_ref_combos = []
        self.required_attributes = []
        self.unwanted_properties = []
        self.load_from_yaml()

    def load_from_yaml(self):
        with open(self.rule_filepath()) as yf:
            rule_ds = yaml.safe_load(yf)

        self.id = rule_ds["id"]

        if rule_ds.get("unwanted_evidence_codes") is not None:
            self.unwanted_evidence_codes = rule_ds["unwanted_evidence_codes"]

        if rule_ds.get("unwanted_evi_code_ref_combos") is not None:
            self.unwanted_evi_code_ref_combos = rule_ds["unwanted_evi_code_ref_combos"]

        if rule_ds.get("required_attributes") is not None:
            self.required_attributes = rule_ds["required_attributes"]
        if self.mod_id():
            self.required_attributes.append({"provided_by": [self.mod_id()]})

        if rule_ds.get("unwanted_properties") is not None:
            self.unwanted_properties = rule_ds["unwanted_properties"]

    @abstractmethod
    def mod_id(self):
        pass

    @abstractmethod
    def rule_filepath(self):
        pass


class DefaultFilterRule(FilterRule):

    def mod_id(self):
        return None

    def rule_filepath(self):
        return "metadata/filter_rules/default.yaml"


class WBFilterRule(FilterRule):
    def mod_id(self):
        return "WB"

    def rule_filepath(self):
        return "metadata/filter_rules/wb.yaml"


class MGIFilterRule(FilterRule):
    def mod_id(self):
        return "MGI"

    def rule_filepath(self):
        return "metadata/filter_rules/mgi.yaml"


def get_filter_rule(mod_id):
    mod_filter_map = {
        'WB': WBFilterRule,
        'MGI': MGIFilterRule
    }
    if mod_id in mod_filter_map:
        return mod_filter_map[mod_id]()
    else:
        return DefaultFilterRule()


class AssocFilter:
    def __init__(self, filter_rule : FilterRule):
        self.filter_rule = filter_rule
        self.ecomap = EcoMap()

    def validate_line(self, assoc):
        evi_code = self.ecomap.ecoclass_to_coderef(assoc["evidence"]["type"])[0]
        if evi_code in self.filter_rule.unwanted_evidence_codes:
            return False
        if len(self.filter_rule.unwanted_evi_code_ref_combos) > 0:
            references = assoc["evidence"]["has_supporting_reference"]
            for evi_ref_combo in self.filter_rule.unwanted_evi_code_ref_combos:
                if evi_ref_combo[0] == evi_code and evi_ref_combo[1] in references:
                    return False
        if len(self.filter_rule.unwanted_properties) > 0 and "annotation_properties" in assoc:
            for up in self.filter_rule.unwanted_properties:
                if up in assoc["annotation_properties"]:
                    return False
        if len(self.filter_rule.required_attributes) > 0:
            meets_requirement = False
            for attr in self.filter_rule.required_attributes:
                # a[attr] is dict
                for k in attr.keys():
                    if assoc[k] in attr[k]:
                        meets_requirement = True
            if not meets_requirement:
                return False
        return True


import logging
from typing import List
from ontobio.io.gpadparser import GpadParser
from ontobio.model.association import GoAssociation
from ontobio.rdfgen.gocamgen import errors

logger = logging.getLogger(__name__)

GPAD_PARSER = GpadParser()
BINDING_ROOT = "GO:0005488"  # binding
IPI_ECO_CODE = "ECO:0000353"


class GoAssocWithFrom:
    """
    Separate with/from column values into
    header vs line arrangement.
    Used for explicit placement in
    annotation assertions.
    """
    def __init__(self, header=None, line=None):
        if header is None:
            header = []
        if line is None:
            line = []
        self.header = sorted(header)
        self.line = sorted(line)

    def add_to_header(self, entity):
        self.header.append(entity)

    def add_to_line(self, entity):
        self.line.append(entity)

    def __str__(self):
        return f"Header: {','.join(self.header)} - Line: {','.join(self.line)}"


class CollapsedAssociationSet:
    def __init__(self, ontology, gpi_entities):
        self.collapsed_associations = []
        self.assoc_dict = {}
        self.go_ontology = ontology
        self.gpi_entities = gpi_entities

    def collapse_annotations(self, associations: List[GoAssociation]):
        # Here we shall decide the distinct assertion instances going into the model
        # This will reduce/eliminate need to SPARQL model graph
        # Group by:
        # 		1. ID
        # 		2. qualifiers (normalize order; any array gotta do this)
        # 		3. primary term
        # 		4. With/From (if primary term is BINDING_ROOT or descendant)
        # 		5. Extensions
        # 	Collapse multiple:
        # 		1. Reference
        # 		2. Evidence Code
        #       3. With/From (if primary term is not BINDING_ROOT or descendant)
        # 		4. Source line
        # 		5. Date
        # 		6. Assigned by
        # 		7. Properties
        for a in associations:
            separated_with_froms: List[GoAssocWithFrom] = self.get_with_froms(a)
            cas = []
            for wf in separated_with_froms:
                ca = self.find_or_create_collapsed_association(a, wf)
                association_line = CollapsedAssociationLine(a, wf.line)
                ca.lines.append(association_line)
                cas.append(ca)

    def find_or_create_collapsed_association(self, association: GoAssociation, with_from: GoAssocWithFrom):
        ca = self.find_by_go_association(association, with_from)
        if ca:
            return ca
        else:
            ca = CollapsedAssociation(association, with_from)
            self.collapsed_associations.append(ca)
            return ca

    def find_by_go_association(self, association: GoAssociation, with_from: GoAssocWithFrom):
        for ca in self.collapsed_associations:
            if ca.header_data_matches(association, with_from):
                return ca

    def __iter__(self):
        return iter(self.collapsed_associations)

    def get_with_froms(self, annot: GoAssociation) -> List[GoAssocWithFrom]:
        # The craziest example:
        # MGI:1  GO:0005515  IPI  PR:1,PR:2|WEIRD:1,PR:3|MGI:2
        with_from_ds = annot.evidence.with_support_from

        # Now arrange these into "header" and "line" values
        eco_code = str(annot.evidence.type)
        term = str(annot.object.id)
        is_binding = eco_code == IPI_ECO_CODE and BINDING_ROOT in self.go_ontology.ancestors(term, reflexive=True)
        if is_binding:
            # Using GPI, check with_froms for taxon equivalency to subj_id
            if self.gpi_entities:
                subject_id = str(annot.subject.id)
                if subject_id in self.gpi_entities:
                     subject_entity = self.gpi_entities[subject_id]
                else:
                    error_message = "Annotation Object ID '{}' missing from provided GPI. Skipping annotation translation.".format(subject_id)
                    logger.warning(error_message)
                    # Throw Exception and except-skip in model builder
                    raise errors.GocamgenException(error_message)

                values_separated: List[GoAssocWithFrom] = []
                for wf in with_from_ds:
                    wf_separated = GoAssocWithFrom()
                    for wf_id in wf.elements:
                        wf_id_str = str(wf_id)
                        wf_entity = self.gpi_entities.get(wf_id_str)
                        if wf_entity and wf_entity.get("taxon") == subject_entity["taxon"]:
                            wf_separated.add_to_header(wf_id_str)
                        wf_separated.add_to_line(wf_id_str)
                    values_separated.append(wf_separated)
            else:
                # Everything is defaulted to header if no GPI available
                values_separated = [GoAssocWithFrom(header=[str(ele) for ele in wf.elements]) for wf in with_from_ds]
                # values_separated = [GoAssocWithFrom(line=[str(ele) for ele in wf.elements]) for wf in with_from_ds]
        else:
            # Everything is defaulted to line if not binding
            values_separated = [GoAssocWithFrom(line=[str(ele) for ele in wf.elements]) for wf in with_from_ds]
        if len(values_separated) == 0:
            # Return empty object if with/from is blank
            values_separated = [GoAssocWithFrom()]
        return values_separated


class CollapsedAssociation:
    def __init__(self, association: GoAssociation, with_from: GoAssocWithFrom):
        self.subject = association.subject
        self.object = association.object
        self.negated = association.negated
        self.qualifiers = sorted(association.qualifiers, key=lambda q: str(q))
        self.with_froms = sorted(with_from.header)
        self.object_extensions = association.object_extensions
        self.lines: List[CollapsedAssociationLine] = []

    def header_data_matches(self, association: GoAssociation, with_from: GoAssocWithFrom):
        if self.subject_id() == str(association.subject.id) and \
           self.object_id() == str(association.object.id) and \
           self.negated == association.negated and \
           self.qualifiers == sorted(association.qualifiers, key=lambda q: str(q)) and \
           self.object_extensions == association.object_extensions and \
           self.with_froms == sorted(with_from.header):
            return True
        return False

    def subject_id(self):
        return str(self.subject.id)

    def object_id(self):
        return str(self.object.id)

    def annot_extensions(self):
        return self.object_extensions

    def with_from(self):
        return self.with_froms

    def __str__(self):
        # TODO: Reconstruct GPAD format or original line - could mean multiple lines for each evidence
        return "{} - {}".format(self.subject_id(), self.object_id())

    def __iter__(self):
        return iter(self.lines)


def dedupe_extensions(extensions):
    new_extensions = []
    for i in extensions:
        if i not in new_extensions:
            new_extensions.append(i)
    return new_extensions


class CollapsedAssociationLine:
    def __init__(self, assoc: GoAssociation, with_from=None):
        self.source_line = assoc.source_line
        self.references = [str(ref) for ref in sorted(assoc.evidence.has_supporting_reference, key=lambda x: str(x))]
        self.evidence_code = str(assoc.evidence.type)
        self.date = assoc.date
        self.assigned_by = assoc.provided_by
        self.with_from = with_from
        self.annotation_properties = extract_properties(assoc)

    def as_dict(self):
        ds = {
            "source_line": self.source_line,
            "evidence": {
                "type": self.evidence_code,
                "has_supporting_reference": self.references
            },
            "date": self.date,
            "provided_by": self.assigned_by,
        }
        if self.annotation_properties:
            ds["annotation_properties"] = self.annotation_properties
        if self.with_from:
            ds["evidence"]["with_support_from"] = self.with_from
        return ds


class CollapsedAssocGocamgenException(errors.GocamgenException):
    def __init__(self, message: str, assoc: CollapsedAssociation):
        self.message = message
        self.assoc = assoc

    def __str__(self):
        return "{}\n{}".format(self.message, "\n".join([l.source_line for l in self.assoc.lines]))


def get_annot_extensions(annot: GoAssociation):
    return annot.object_extensions


def extract_properties_from_string(prop_col):
    props = prop_col.split("|")
    props_dict = {}
    for p in props:
        k, v = p.split("=")
        if k in props_dict:
            props_dict[k].append(v)
        else:
            props_dict[k] = [v]
    return props_dict


def extract_properties(annot: GoAssociation):
    annotation_properties = {}
    property_keys = set([prop[0] for prop in annot.properties])
    for pk in property_keys:
        annotation_properties[pk] = annot.annotation_property_values(pk)
    return annotation_properties



import io

from ontobio.io import entitywriter, gafgpibridge


def test_header_newline():
    gpi_obj = {
        'id': "MGI:MGI:1918911",
        'label': "0610005C13Rik",  # db_object_symbol,
        'full_name': "RIKEN cDNA 0610005C13 gene",  # db_object_name,
        'synonyms': [],
        'type': ["gene"],  # db_object_type,
        'parents': "",  # GAF does not have this field, but it's optional in GPI
        'xrefs': "",  # GAF does not have this field, but it's optional in GPI
        'taxon': {
            'id': "NCBITaxon:10090"
        }
    }
    entity = gafgpibridge.Entity(gpi_obj)

    out = io.StringIO()
    gpiwriter = entitywriter.GpiWriter(file=out)
    gpiwriter.write_entity(entity)
    outlines = out.getvalue().split("\n")

    expected_lines = [
        "!gpi-version: 1.2",
        "MGI\tMGI:1918911\t0610005C13Rik\tRIKEN cDNA 0610005C13 gene\t\tgene\ttaxon:10090\t\t\t",
        ""
    ]
    assert expected_lines == outlines



from ontobio.io import assocparser
from ontobio.io import entityparser
from ontobio.io import entitywriter
from ontobio.io.assocparser import ENTITY, EXTENSION, ANNOTATION, Report
from ontobio.io import parser_version_regex
from ontobio.io import qc
from ontobio.model import association, collections

from ontobio.rdfgen import relations
from prefixcommons import curie_util

from typing import List, Dict, Optional
from dataclasses import dataclass

import functools
import re
import logging

logger = logging.getLogger(__name__)


gpad_line_validators = {
    "default": assocparser.ColumnValidator(),
    "qualifier2_1": assocparser.Qualifier2_1(),
    "qualifier2_2": assocparser.Qualifier2_2(),
    "curie": assocparser.CurieValidator(),
    "taxon": assocparser.TaxonValidator(),
}

class GpadParser(assocparser.AssocParser):
    """
    Parser for GO GPAD Format

    https://github.com/geneontology/go-annotation/blob/master/specs/gpad-gpi-1_2.md
    """


    def __init__(self, config=assocparser.AssocParserConfig(), group="unknown", dataset="unknown", bio_entities=None):
        """
        Arguments:
        ---------

        config : a AssocParserConfig object
        """
        self.config = config
        self.report = assocparser.Report(config=self.config, group="unknown", dataset="unknown")
        self.version = None
        self.default_version = "1.2"
        self.bio_entities = bio_entities
        if self.bio_entities is None:
            self.bio_entities = collections.BioEntities(dict())
        if self.config.gpi_authority_path is not None:
            gpi_paths = self.config.gpi_authority_path
            if isinstance(gpi_paths, str):
                gpi_paths = [gpi_paths]
            for gpi_path in gpi_paths:
                gpi_bio_entities = collections.BioEntities.load_from_file(gpi_path)
                self.bio_entities.merge(gpi_bio_entities)
                print("Loaded {} entities from {}".format(len(gpi_bio_entities.entities.keys()), gpi_path))
        # self.gpi = dict()
        # if self.config.gpi_authority_path is not None:
        #     print("Loading GPI...")
        #     self.gpi = dict()
        #     parser = entityparser.GpiParser()
        #     with open(self.config.gpi_authority_path) as gpi_f:
        #         entities = parser.parse(file=gpi_f)
        #         for entity in entities:
        #             self.gpi[entity["id"]] = {
        #                 "symbol": entity["label"],
        #                 "name": entity["full_name"],
        #                 "synonyms": entitywriter.stringify(entity["synonyms"]),
        #                 "type": entity["type"]
        #             }
        #         print("Loaded {} entities from {}".format(len(self.gpi.keys()), self.config.gpi_authority_path))

    def gpad_version(self) -> str:
        if self.version:
            return self.version
        else:
            return self.default_version

    def skim(self, file):
        file = self._ensure_file(file)
        tuples = []
        for line in file:
            if line.startswith("!"):
                continue
            vals = line.split("\t")
            if len(vals) != 12:
                logger.error("Unexpected number of columns: {}. GPAD should have 12.".format(vals))
            rel = vals[2]

            negated, relation, _ = self._parse_qualifier(vals[2], None)

            # never include NOTs in a skim
            if negated:
                continue
            if self._is_exclude_relation(relation):
                continue


            id = self._pair_to_id(vals[0], vals[1])
            if not self._validate_id(id, line, context=ENTITY):
                continue
            t = vals[3]
            tuples.append( (id,None,t) )
        return tuples

    def parse_line(self, line):
        """Parses a single line of a GPAD.

        Return a tuple `(processed_line, associations)`. Typically
        there will be a single association, but in some cases there
        may be none (invalid line) or multiple (disjunctive clause in
        annotation extensions)

        Note: most applications will only need to call this directly if they require fine-grained control of parsing. For most purposes,
        :method:`parse_file` can be used over the whole file

        Arguments
        ---------
        line : str
            A single tab-seperated line from a GPAD file

        """
        parsed = super().validate_line(line)
        if parsed:
            return parsed

        if self.is_header(line):
            if self.version is None:
                # We are still looking
                parsed = parser_version_regex.findall(line)
                if len(parsed) == 1:
                    filetype, version, _ = parsed[0]
                    if version == "2.0":
                        logger.info("Detected GPAD version 2.0")
                        self.version = version
                    else:
                        logger.info("Detected GPAD version {}, so defaulting to 1.2".format(version))
                        self.version = self.default_version

            return assocparser.ParseResult(line, [{ "header": True, "line": line.strip() }], False)

        # At this point, we should have gone through all the header, and a version number should be established
        if self.version is None:
            logger.warning("No version number found for this file so we will assume GPAD version: {}".format(self.default_version))
            self.version = self.default_version

        vals = [el.strip() for el in line.split("\t")]

        parsed = to_association(list(vals), report=self.report, version=self.gpad_version(), bio_entities=self.bio_entities)
        if parsed.associations == []:
            return parsed

        assoc = parsed.associations[0]

        split_line = assocparser.SplitLine(line=line, values=vals, taxon="")

        valid_goid = self._validate_ontology_class_id(str(assoc.object.id), split_line)
        if valid_goid is None:
            return assocparser.ParseResult(line, [], True)
        assoc.object.id = association.Curie.from_str(valid_goid)

        go_rule_results = qc.test_go_rules(assoc, self.config)
        for rule, result in go_rule_results.all_results.items():
            if result.result_type == qc.ResultType.WARNING:
                self.report.warning(line, assocparser.Report.VIOLATES_GO_RULE, "",
                                    msg="{id}: {message}".format(id=rule.id, message=result.message), rule=int(rule.id.split(":")[1]))

            if result.result_type == qc.ResultType.ERROR:
                self.report.error(line, assocparser.Report.VIOLATES_GO_RULE, "",
                                    msg="{id}: {message}".format(id=rule.id, message=result.message), rule=int(rule.id.split(":")[1]))
                # Skip the annotation
                return assocparser.ParseResult(line, [], True)

            if result.result_type == qc.ResultType.PASS:
                self.report.message(assocparser.Report.INFO, line, Report.RULE_PASS, "",
                                    msg="Passing Rule", rule=int(rule.id.split(":")[1]))

        assoc = go_rule_results.annotation  # type: association.GoAssociation

        if not self._validate_id(str(assoc.subject.id), split_line, context=ENTITY):
            return assocparser.ParseResult(line, [], True)

        if not self._validate_id(str(assoc.object.id), split_line, context=ANNOTATION):
            return assocparser.ParseResult(line, [], True)

        if not self._validate_id(str(assoc.evidence.type), split_line):
            return assocparser.ParseResult(line, [], True)

        if assoc.interacting_taxon:
            if not self._validate_taxon(str(assoc.interacting_taxon), split_line):
                self.report.error(line, assocparser.Report.INVALID_TAXON, str(assoc.interacting_taxon), "Taxon ID is invalid", rule=27)
                return assocparser.ParseResult(line, [], True)


        #TODO: ecomap is currently one-way only
        #ecomap = self.config.ecomap
        #if ecomap != None:
        #    if ecomap.ecoclass_to_coderef(evidence) == (None,None):
        #        self.report.error(line, Report.UNKNOWN_EVIDENCE_CLASS, evidence,
        #                          msg="Expecting a known ECO class ID")


        # Reference Column
        references = self.validate_curie_ids(assoc.evidence.has_supporting_reference, split_line)
        if references is None:
            return assocparser.ParseResult(line, [], True)

        # With/From
        for wf in assoc.evidence.with_support_from:
            validated = self.validate_curie_ids(wf.elements, split_line)
            if validated is None:
                return assocparser.ParseResult(line, [], True)


        return assocparser.ParseResult(line, [assoc], False)

    def is_header(self, line):
        return line.startswith("!")


relation_tuple = re.compile(r'(.+)\((.+)\)')

def from_1_2(gpad_line: List[str], report=None, group="unknown", dataset="unknown", bio_entities=None):
    source_line = "\t".join(gpad_line)

    if source_line == "":
        report.error(source_line, "Blank Line", "EMPTY", "Blank lines are not allowed", rule=1)
        return assocparser.ParseResult(source_line, [], True, report=report)

    if len(gpad_line) > 12:
        report.warning(source_line, assocparser.Report.WRONG_NUMBER_OF_COLUMNS, "",
            msg="There were more than 12 columns in this line. Proceeding by cutting off extra columns.",
            rule=1)

        gpad_line = gpad_line[:12]

    if 12 > len(gpad_line) >= 10:
        gpad_line += [""] * (12 - len(gpad_line))

    if len(gpad_line) != 12:
        report.error(source_line, assocparser.Report.WRONG_NUMBER_OF_COLUMNS, "",
            msg="There were {columns} columns found in this line, and there should be between 10 and 12".format(columns=len(gpad_line)), rule=1)
        return assocparser.ParseResult(source_line, [], True, report=report)

    ## check for missing columns
    ## We use indeces here because we run GO RULES before we split the vals into individual variables
    DB_INDEX = 0
    DB_OBJECT_INDEX = 1
    QUALIFIER = 2
    REFERENCE_INDEX = 4
    EVIDENCE_INDEX = 5
    if gpad_line[DB_INDEX] == "":
        report.error(source_line, Report.INVALID_IDSPACE, "EMPTY", "col1 is empty", rule=1)
        return assocparser.ParseResult(source_line, [], True, report=report)
    if gpad_line[DB_OBJECT_INDEX] == "":
        report.error(source_line, Report.INVALID_ID, "EMPTY", "col2 is empty", rule=1)
        return assocparser.ParseResult(source_line, [], True, report=report)
    if gpad_line[QUALIFIER] == "":
        report.error(source_line, Report.INVALID_TAXON, "EMPTY", "qualifier column is empty", rule=1)
        return assocparser.ParseResult(source_line, [], True, report=report)
    if gpad_line[REFERENCE_INDEX] == "":
        report.error(source_line, Report.INVALID_ID, "EMPTY", "reference column is empty", rule=1)
        return assocparser.ParseResult(source_line, [], True, report=report)
    if gpad_line[EVIDENCE_INDEX] == "":
        report.error(source_line, Report.INVALID_ID, "EMPTY", "Evidence column is empty", rule=1)

    taxon = association.Curie("NCBITaxon", "0")
    subject_curie = association.Curie(gpad_line[0], gpad_line[1])
    subject = association.Subject(subject_curie, "", [""], [], [], taxon)

    entity = bio_entities.get(subject_curie)
    if entity is not None:
        subject = entity
        taxon = subject.taxon

    go_term = association.Curie.from_str(gpad_line[3])
    if go_term.is_error():
        report.error(source_line, Report.INVALID_SYMBOL, gpad_line[3], "Problem parsing GO Term", taxon=str(taxon), rule=1)
        return assocparser.ParseResult(source_line, [], True, report=report)

    object = association.Term(go_term, taxon)

    evidence_type = association.Curie.from_str(gpad_line[5])
    if evidence_type.is_error():
        report.error(source_line, Report.INVALID_SYMBOL, gpad_line[5], "Problem parsing Evidence ECO Curie", taxon=str(taxon), rule=1)
        return assocparser.ParseResult(source_line, [], True, report=report)

    references = [association.Curie.from_str(e) for e in gpad_line[4].split("|") if e]
    for r in references:
        if r.is_error():
            report.error(source_line, Report.INVALID_SYMBOL, gpad_line[4], "Problem parsing references", taxon=str(taxon), rule=1)
            return assocparser.ParseResult(source_line, [], True, report=report)

    withfroms = association.ConjunctiveSet.str_to_conjunctions(gpad_line[6])  # Returns a list of ConjuctiveSets or Error
    if isinstance(withfroms, association.Error):
        report.error(source_line, Report.INVALID_SYMBOL, gpad_line[6], "Problem parsing With/From column", taxon=str(taxon), rule=1)
        return assocparser.ParseResult(source_line, [], True, report=report)

    evidence = association.Evidence(evidence_type, references, withfroms)

    # Guarenteed to have at least one element, from above check
    raw_qs = gpad_line[QUALIFIER].split("|")
    negated = "NOT" in raw_qs

    looked_up_qualifiers = [relations.lookup_label(q) for q in raw_qs if q != "NOT"]
    if None in looked_up_qualifiers:
        report.error(source_line, Report.INVALID_QUALIFIER, raw_qs, "Could not find a URI for qualifier", taxon=str(taxon), rule=1)
        return assocparser.ParseResult(source_line, [], True, report=report)

    qualifiers = [association.Curie.from_str(curie_util.contract_uri(q)[0]) for q in looked_up_qualifiers]

    date = assocparser.parse_date(gpad_line[8], report, source_line)
    if date is None:
        return assocparser.ParseResult(source_line, [], True, report=report)

    interacting_taxon = None
    if gpad_line[7]:
        taxon_result = gpad_line_validators["taxon"].validate(gpad_line[7])
        if not taxon_result.valid:
            report.error(source_line, Report.INVALID_TAXON, taxon_result.original, taxon_result.message, taxon=str(taxon_result.original), rule=1)
            return assocparser.ParseResult(source_line, [], True, report=report)
        else:
            interacting_taxon = taxon_result.parsed[0]

    conjunctions = []
    if gpad_line[10]:
        conjunctions = association.ConjunctiveSet.str_to_conjunctions(
            gpad_line[10],
            conjunct_element_builder=lambda el: association.ExtensionUnit.from_str(el))

        if isinstance(conjunctions, association.Error):
            report.error(source_line, Report.EXTENSION_SYNTAX_ERROR, conjunctions.info, "extensions should be relation(curie)", taxon=str(taxon), rule=1)
            return assocparser.ParseResult(source_line, [], True, report=report)

    properties_list = association.parse_annotation_properties(gpad_line[11])


    # print(properties_list)
    a = association.GoAssociation(
        source_line=source_line,
        subject=subject,
        relation=qualifiers[0],
        object=object,
        negated=negated,
        qualifiers=qualifiers,
        aspect=None,
        interacting_taxon=interacting_taxon,
        evidence=evidence,
        subject_extensions=[],
        object_extensions=conjunctions,
        provided_by=gpad_line[9],
        date=date,
        properties=properties_list)

    return assocparser.ParseResult(source_line, [a], False, report=report)

def from_2_0(gpad_line: List[str], report=None, group="unknown", dataset="unknown", bio_entities=None):
    source_line = "\t".join(gpad_line)

    if source_line == "":
        report.error(source_line, "Blank Line", "EMPTY", "Blank lines are not allowed", rule=1)
        return assocparser.ParseResult(source_line, [], True, report=report)

    if len(gpad_line) > 12:
        report.warning(source_line, assocparser.Report.WRONG_NUMBER_OF_COLUMNS, "",
            msg="There were more than 12 columns in this line. Proceeding by cutting off extra columns.",
            rule=1)

        gpad_line = gpad_line[:12]

    if 12 > len(gpad_line) >= 10:
        gpad_line += [""] * (12 - len(gpad_line))

    if len(gpad_line) != 12:
        report.error(source_line, assocparser.Report.WRONG_NUMBER_OF_COLUMNS, "",
            msg="There were {columns} columns found in this line, and there should be between 10 and 12".format(columns=len(gpad_line)), rule=1)
        return assocparser.ParseResult(source_line, [], True, report=report)

    ## check for missing columns
    ## We use indeces here because we run GO RULES before we split the vals into individual variables
    SUBJECT_CURIE = 0
    RELATION = 2
    ONTOLOGY_CLASS_INDEX = 3
    REFERENCE_INDEX = 4
    EVIDENCE_INDEX = 5
    DATE_INDEX = 8
    ASSIGNED_BY_INDEX = 9
    required = [SUBJECT_CURIE, RELATION, ONTOLOGY_CLASS_INDEX, REFERENCE_INDEX, EVIDENCE_INDEX, DATE_INDEX, ASSIGNED_BY_INDEX]
    for req in required:
        if gpad_line[req] == "":
            report.error(source_line, Report.INVALID_ID, "EMPTY", "Column {} is empty".format(req + 1), rule=1)
            return assocparser.ParseResult(source_line, [], True, report=report)

    taxon = association.Curie("NCBITaxon", "0")
    subject_curie = association.Curie.from_str(gpad_line[SUBJECT_CURIE])
    if subject_curie.is_error():
        report.error(source_line, Report.INVALID_SYMBOL, gpad_line[SUBJECT_CURIE], "Problem parsing DB Object", taxon=str(taxon), rule=1)
        return assocparser.ParseResult(source_line, [], True, report=report)

    subject = association.Subject(subject_curie, "", "", [], "", taxon)
    entity = bio_entities.get(subject_curie)
    if entity is not None:
        # If we found a subject entity, then set `subject` to the found entity
        subject = entity
        taxon = subject.taxon

    negated = gpad_line[1] == "NOT"

    relation = association.Curie.from_str(gpad_line[RELATION])
    if relation.is_error():
        report.error(source_line, Report.INVALID_SYMBOL, gpad_line[RELATION], "Problem parsing Relation", taxon=str(taxon), rule=1)
        return assocparser.ParseResult(source_line, [], True, report=report)

    go_term = association.Curie.from_str(gpad_line[ONTOLOGY_CLASS_INDEX])
    if go_term.is_error():
        report.error(source_line, Report.INVALID_SYMBOL, gpad_line[ONTOLOGY_CLASS_INDEX], "Problem parsing GO Term", taxon=str(taxon), rule=1)
        return assocparser.ParseResult(source_line, [], True, report=report)

    object = association.Term(go_term, taxon)

    evidence_type = association.Curie.from_str(gpad_line[EVIDENCE_INDEX])
    if evidence_type.is_error():
        report.error(source_line, Report.INVALID_SYMBOL, gpad_line[EVIDENCE_INDEX], "Problem parsing Evidence ECO Curie", taxon=str(taxon), rule=1)
        return assocparser.ParseResult(source_line, [], True, report=report)

    references = [association.Curie.from_str(e) for e in gpad_line[REFERENCE_INDEX].split("|") if e]
    for r in references:
        if r.is_error():
            report.error(source_line, Report.INVALID_SYMBOL, gpad_line[REFERENCE_INDEX], "Problem parsing references", taxon=str(taxon), rule=1)
            return assocparser.ParseResult(source_line, [], True, report=report)

    withfroms = association.ConjunctiveSet.str_to_conjunctions(gpad_line[6])  # Returns a list of ConjuctiveSets or Error
    if isinstance(withfroms, association.Error):
        report.error(source_line, Report.INVALID_SYMBOL, gpad_line[6], "Problem parsing With/From column", taxon=str(taxon), rule=1)
        return assocparser.ParseResult(source_line, [], True, report=report)

    evidence = association.Evidence(evidence_type, references, withfroms)

    interacting_taxon = None
    if gpad_line[7] != "":
        interacting_taxon = association.Curie.from_str(gpad_line[7])
        if interacting_taxon.is_error():
            report.error(source_line, Report.INVALID_SYMBOL, gpad_line[7], "Problem parsing Interacting Taxon", taxon=str(taxon), rule=1)
            return assocparser.ParseResult(source_line, [], True, report=report)

    date = assocparser.parse_iso_date(gpad_line[DATE_INDEX], report, source_line)
    if date is None:
        return assocparser.ParseResult(source_line, [], True, report=report)

    conjunctions = []
    # The elements of the extension units are Curie(Curie)
    if gpad_line[10]:
        conjunctions = association.ConjunctiveSet.str_to_conjunctions(
            gpad_line[10],
            conjunct_element_builder=lambda el: association.ExtensionUnit.from_curie_str(el))

        if isinstance(conjunctions, association.Error):
            report.error(source_line, Report.EXTENSION_SYNTAX_ERROR, conjunctions.info, "extensions should be relation(curie)", taxon=str(taxon), rule=1)
            return assocparser.ParseResult(source_line, [], True, report=report)

    properties_list = association.parse_annotation_properties(gpad_line[11])

    a = association.GoAssociation(
        source_line=source_line,
        subject=subject,
        relation=relation,
        object=object,
        negated=negated,
        qualifiers=[relation],
        aspect=None,
        interacting_taxon=interacting_taxon,
        evidence=evidence,
        subject_extensions=[],
        object_extensions=conjunctions,
        provided_by=gpad_line[9],
        date=date,
        properties=properties_list)

    return assocparser.ParseResult(source_line, [a], False, report=report)

def to_association(gpad_line: List[str], report=None, group="unknown", dataset="unknown", version="1.2", bio_entities=None) -> assocparser.ParseResult:
    report = Report(group=group, dataset=dataset) if report is None else report
    bio_entities = collections.BioEntities(dict()) if bio_entities is None else bio_entities
    if version == "2.0":
        return from_2_0(gpad_line, report=report, group=group, dataset=dataset, bio_entities=bio_entities)
    else:
        return from_1_2(gpad_line, report=report, group=group, dataset=dataset, bio_entities=bio_entities)



from ontobio.sim.api.interfaces import SimApi
from typing import Iterable, Collection, Union, Optional, List
from ontobio.model.similarity import SimResult, TypedNode
from ontobio.vocabulary.similarity import SimAlgorithm
from ontobio.sim.api.interfaces import FilteredSearchable
from ontobio.golr.golr_associations import get_objects_for_subject
from ontobio.util.scigraph_util import get_id_type_map, typed_node_from_id


class PhenoSimEngine():
    """
    Engine for computing phenotype similarity

    A wrapper for a similarity API whereby non-phenotype identifiers
    are resolved to a list of phenotypes
    """

    def __init__(self, sim_api: SimApi):
        self.sim_api = sim_api

    def search(
            self,
            id_list: Iterable[str],
            negated_ids: Optional[Collection] = None,
            limit: Optional[int] = 100,
            taxon_filter: Optional[str]= None,
            category_filter: Optional[str]= None,
            method: Optional[SimAlgorithm] = SimAlgorithm.PHENODIGM,
            is_feature_set: bool = True
    ) -> SimResult:
        """
        Execute a search using sim_api,

        :param taxon_filter: only
        :param category_filter:
        :param method: similarity metric
        :param is_feature_set: if True, treats all identifiers as phenotypic
                               features, else resolves ids to a list of phenotypes
                               (eg set to false if some identifiers are
                                diseases, genes, etc)

        :raises NotImplementedError:
            - If sim method or filters are not supported
        """

        if negated_ids is None: negated_ids = []

        if method not in self.sim_api.matchers():
            raise NotImplementedError("Sim method not implemented "
                                      "in {}".format(str(self.sim_api)))

        if not is_feature_set:
            # Determine if entity is a phenotype or individual containing
            # a pheno profile (gene, disease, case, etc)
            pheno_list = PhenoSimEngine._resolve_nodes_to_phenotypes(id_list)
        else:
            pheno_list = id_list

        if taxon_filter is not None or category_filter is not None:
            if not isinstance(self.sim_api, FilteredSearchable):
                raise NotImplementedError("filtered search not implemented "
                                          "in {}".format(str(self.sim_api)))
            search_result = self.sim_api.filtered_search(
                pheno_list, negated_ids, limit, taxon_filter, category_filter, method
            )
        else:
            search_result = self.sim_api.search(pheno_list, negated_ids, limit, method)

        return search_result

    def compare(self,
                reference_ids: Collection,
                query_profiles: Collection[Collection],
                method: Optional[SimAlgorithm] = SimAlgorithm.PHENODIGM,
                is_feature_set: bool = True) -> SimResult:
        """
        Execute one or more comparisons using sim_api
        :param reference_ids: a list of phenotypes or ids that comprise
                              one or more phenotypes
        :param query_profiles: a list of lists of phenotypes or ids
                                   that comprise one or more phenotypes
        :param method: similarity metric
        :param is_feature_set: if True, treats all identifiers as phenotypic
                               features, else resolves ids to a list of phenotypes
                               (eg set to false if some identifiers are
                                diseases, genes, etc)
        :return: SimResult object
        :raises NotImplementedError: If sim method or filters are not supported
        """
        if method not in self.sim_api.matchers():
            raise NotImplementedError("Sim method not implemented "
                                      "in {}".format(str(self.sim_api)))

        is_first_result = True
        comparisons = None

        if is_feature_set:
            reference_phenos = reference_ids
        else:
            reference_phenos = PhenoSimEngine._resolve_nodes_to_phenotypes(reference_ids)
        for query_profile in query_profiles:
            if is_feature_set:
                query_phenos = query_profile
            else:
                query_phenos = PhenoSimEngine._resolve_nodes_to_phenotypes(query_profile)

            sim_result = self.sim_api.compare(reference_phenos, query_phenos, method)
            if len(query_profile) > 1:
                id = " + ".join(query_profile)
                sim_result.matches[0].id = id
                sim_result.matches[0].label = id
            else:
                node = typed_node_from_id(query_profile[0])
                if sim_result.matches:
                    sim_result.matches[0].id = node.id
                    sim_result.matches[0].label = node.label
                    sim_result.matches[0].type = node.type
                    sim_result.matches[0].taxon = node.taxon

            if is_first_result:
                comparisons = sim_result
                is_first_result = False
            else:
                comparisons.matches.append(sim_result.matches[0])
                comparisons.query.target_ids.append(sim_result.query.target_ids[0])

        if len(reference_ids) == 1:
            comparisons.query.reference = typed_node_from_id(reference_ids[0])
        else:
            reference_id = " + ".join(reference_ids)
            comparisons.query.reference = TypedNode(
                id=reference_id,
                label=reference_id,
                type='unknown'
            )

        return comparisons

    @staticmethod
    def _resolve_nodes_to_phenotypes(id_list: Iterable[str]) -> Iterable[str]:
        """
        Given a list of ids of unknown type, determine which ids
        are phenotypes, if the id is not a phenotype, check to
        see if it is associated with one or more phenotypes via
        the 'has_phenotype' relation
        :param id_list: list of ids of any type (curies as strings)
        :return: list of phenotypes (curies as strings)
        """
        pheno_list = []
        node_types = get_id_type_map(id_list)

        for node in id_list:
            if 'phenotype' in node_types[node]:
                pheno_list.append(node)
            else:
                phenotypes = get_objects_for_subject(
                    subject=node, object_category='phenotype', relation='RO:0002200'
                )
                pheno_list = pheno_list + phenotypes
        return pheno_list

    @staticmethod
    def _get_node_info(id_list: Iterable[str]) -> Iterable[str]:
        """
        Given a list of ids of unknown type, determine which ids
        are phenotypes, if the id is not a phenotype, check to
        see if it is associated with one or more phenotypes via
        the 'has_phenotype' relation
        :param id_list: list of ids of any type (curies as strings)
        :return: list of phenotypes (curies as strings)
        """
        pheno_list = []
        node_types = get_id_type_map(id_list)

        for node in id_list:
            if 'phenotype' in node_types[node]:
                pheno_list.append(node)
            else:
                phenotypes = get_objects_for_subject(
                    subject=node, object_category='phenotype', relation='RO:0002200'
                )
                pheno_list = pheno_list + phenotypes
        return pheno_list



"""
This contains the data model for parsing annotations from GAF and GPAD.

The idea is to make it easy to parse text lines of any source into a GoAssociation object
and then give the GoAssociation object the ability to convert itself into GPAD or GAF 
of any version. Or any other format that is required.


"""

import json
import typing
import collections
import enum
import datetime
import re
import logging

import bidict
from prefixcommons import curie_util

from ontobio.rdfgen import relations
from ontobio.ecomap import EcoMap

ecomap = EcoMap()
ecomap.mappings()


from typing import List, Optional, NamedTuple, Dict, Callable, Union, TypeVar, Tuple
from dataclasses import dataclass

logger = logging.getLogger(__name__)


Aspect = typing.NewType("Aspect", str)
Provider = typing.NewType("Provider", str)
Date = collections.namedtuple("Date", ["year", "month", "day", "time"])


def ymd_str(date: Date, separator: str) -> str:
    return "{year}{sep}{month}{sep}{day}".format(year=date.year, sep=separator, month=date.month, day=date.day)


@dataclass
class Error:
    info: str
    entity: str = ""

    def is_error(self):
        return True

def TwoTupleStr(items: List[str]) -> tuple:
    """
    Create a tuple of of str that is guaranteed to be of length two from a list

    If the list is larger, then only the first two elements will be used.
    If the list is smaller, then the empty string will be used
    """

    if len(items) > 2:
        items = items[:2]
    
    if len(items) < 2:
        items += [""] * (2 - len(items))
    
    return tuple([str(i) for i in items])


def parse_annotation_properties(properties_field: str) -> List[tuple]:
    properties_list = [TwoTupleStr(prop.split("=", maxsplit=1)) for prop in properties_field.split("|") if prop]
    return properties_list

@dataclass(unsafe_hash=True)
class Curie:
    """
    Object representing a Compact URI, with a namespace identifier along with an ID, like GO:1234567.

    Use `from_str` to parse a string like "GO:1234567" into a Curie. The result should be checked for errors
    with `is_error`
    """
    namespace: str
    identity: str

    def __str__(self) -> str:
        return "{}:{}".format(self.namespace, self.identity)

    @classmethod
    def from_str(Curie, entity: str):
        splitup = entity.split(":", maxsplit=1)
        splitup += [""] * (2 - len(splitup))
        namespace, identity = splitup
        if namespace == "" and identity == "":
            return Error("Namespace and Identity of CURIE is empty")

        if namespace == "":
            return Error("Namespace of CURIE is empty")

        if identity == "":
            return Error("Identity of CURIE is empty")

        if " " in namespace or " " in identity:
            return Error("No spaces allowed in CURIEs")

        return Curie(namespace, identity)

    def is_error(self) -> bool:
        return False


@dataclass(unsafe_hash=True)
class Subject:
    id: Curie
    label: str
    """
    label is also `DB_Object_Symbol` in the GPI spec
    """

    fullname: List[str]
    """
    fullname is also `DB_Object_Name` in the GPI spec, cardinality 0+
    """

    synonyms: List[str]
    """
    Cardinality 0+
    """

    type: List[Curie]
    """
    In GPI 1.2, this was a string, corresponding to labels of the Sequence Ontology
    gene, protein_complex; protein; transcript; ncRNA; rRNA; tRNA; snRNA; snoRNA,
    any subclass of ncRNA.
    If the specific type is unknown, use `gene_product`.

    When reading gpi 1.2, these labels should be mapped to the 2.0 spec, stating that
    the type must be a Curie in the Sequence Ontology OR Protein Ontology OR Gene Ontology

    In GPI 1.2, there is only 1 value, and is required.
    In GPI 2.0 there is a minimum of 1, but maybe more.

    If writing out to GPI 1.2/GAF just take the first value in the list.
    """

    taxon: Curie
    """
    Should be NCBITaxon:...
    """

    encoded_by: List[Curie]
    """
    Optional, or cardinality 0+
    """

    parents: List[Curie]
    """
    Optional, or cardinality 0+
    """

    contained_complex_members: List[Curie]
    """
    Optional, or cardinality 0+
    """

    db_xrefs: List[Curie]
    """
    Optional, or cardinality 0+
    """

    properties: Dict[str, str]
    """
    Optional, or cardinality 0+
    """

    def __init__(self, id: Curie, label: str, fullname: List[str], synonyms: List[str], type: Union[List[str], List[Curie]], taxon: Curie,
                    encoded_by: List[Curie]=None, parents: List[Curie]=None, contained_complex_members: List[Curie]=None,
                    db_xrefs: List[Curie]=None, properties: Dict=None):

        if len(type) > 0:
            if isinstance(type[0], str):
                # If the incoming `type` list is strings, then they are labels, so convert to Curie
                self.type = [map_gp_type_label_to_curie(t) for t in type]
            else:
                # We will assume that incoming type is a Curie at this point, if not a string
                self.type = type
        else:
            # If we didn't receive anything, then default to "gene_produce"
            self.type = [Curie(namespace="CHEBI", identity="33695")]

        self.id = id
        self.label = label
        self.fullname = fullname
        self.synonyms = synonyms
        self.taxon = taxon
        self.encoded_by = encoded_by if encoded_by else []
        self.parents = parents if parents else []
        self.contained_complex_members = contained_complex_members if contained_complex_members else []
        self.db_xrefs = db_xrefs if db_xrefs else []
        self.properties = properties if properties else dict()

    def fullname_field(self, max=None) -> str:
        """
        Converts the `fullname` or `DB_Object_Name` into the field text string used in files
        """

        if not max:
            max = len(self.fullname)

        return "|".join([n for n in self.fullname[0:max]])

# ===============================================================================
__default_entity_type_to_curie_mapping = bidict.bidict({
    "protein_coding_gene": Curie.from_str("SO:0001217"),
    "snRNA": Curie.from_str("SO:0000274"),
    "ncRNA": Curie.from_str("SO:0000655"),
    "rRNA": Curie.from_str("SO:0000252"),
    "mRNA": Curie.from_str("SO:0000234"),
    "lnc_RNA": Curie.from_str("SO:0001877"),
    "lincRNA": Curie.from_str("SO:0001463"),
    "tRNA": Curie.from_str("SO:0000253"),
    "snoRNA": Curie.from_str("SO:0000275"),
    "miRNA": Curie.from_str("SO:0000276"),
    "RNA": Curie.from_str("SO:0000356"),
    "scRNA": Curie.from_str("SO:0000013"),
    "piRNA": Curie.from_str("SO:0001035"),
    "tmRNA": Curie.from_str("SO:0000584"),
    "SRP_RNA": Curie.from_str("SO:0000590"),
    "primary_transcript": Curie.from_str("SO:0000185"),
    "ribozyme": Curie.from_str("SO:0000374"),
    "telomerase_RNA": Curie.from_str("SO:0000390"),
    "RNase_P_RNA": Curie.from_str("SO:0000386"),
    "antisense_RNA": Curie.from_str("SO:0000644"),
    "RNase_MRP_RNA": Curie.from_str("SO:0000385"),
    "guide_RNA": Curie.from_str("SO:0000602"),
    "hammerhead_ribozyme": Curie.from_str("SO:0000380"),
    "protein": Curie.from_str("PR:000000001"),
    "marker or uncloned locus": Curie.from_str("SO:0001645"),
    "gene segment": Curie.from_str("SO:3000000"),
    "pseudogene": Curie.from_str("SO:0000336"),
    "gene": Curie.from_str("SO:0000704"),
    "biological region": Curie.from_str("SO:0001411"),
    "protein_complex": Curie.from_str("GO:0032991"),
    "transcript": Curie.from_str("SO:0000673"),
    "gene_product": Curie.from_str("CHEBI:33695"),
    "ncRNA-coding gene": Curie.from_str("SO:0001263"),
    "antisense_lncRNA": Curie.from_str("SO:0001904"),
    "transposable_element_gene": Curie.from_str("SO:0000111")
})

def map_gp_type_label_to_curie(type_label: str) -> Curie:
    """
    Map entity types in GAF or GPI 1.2 into CURIEs in Sequence Ontology (SO),
    Protein Ontology (PRO), or Gene Ontology (GO).

    This is a measure to upgrade the pseudo-labels into proper Curies. Present here are
    the existing set of labels in current use, and how they should be mapped into CURIEs.
    """
    # normalized_label = type_label.translate()
    global __default_entity_type_to_curie_mapping
    return __default_entity_type_to_curie_mapping.get(type_label, __default_entity_type_to_curie_mapping["gene_product"])

def gp_type_label_to_curie(type: Curie) -> str:
    """
    This is the reverse of `map_gp_type_label_to_curie`
    """
    global __default_entity_type_to_curie_mapping
    return __default_entity_type_to_curie_mapping.inverse.get(type, "gene_product")

@dataclass(unsafe_hash=True)
class Term:
    """
    Represents a Gene Ontology term
    """
    id: Curie
    taxon: Curie

C = TypeVar("C")


@dataclass(unsafe_hash=True)
class ConjunctiveSet:
    """
    This respresents a comma separated list of objects which can be turned into strings.

    This is used for the with/from and extensions fields in the GoAssociation.

    The field `elements` can be a list of Curie or ExtensionUnit. Curie for with/from, and
    ExtensionUnit for extensions field.
    """
    elements: List

    def __str__(self) -> str:
        return ",".join([str(conj) for conj in self.elements])

    def display(self, conjunct_to_str=lambda c: str(c)) -> str:
        """
        Convert this ConjunctiveSet to a string separateted by commas.

        This calls `conjunct_to_str` (which defaults to `str`) on each element before joining. To use a different
        string representation of each element, pass in a different function. This functionality is used to differentiate
        between GPAD 1.2 and GPAD 2.0, where relations are written differently per version.
        """
        return ",".join([conjunct_to_str(conj) for conj in self.elements])

    @classmethod
    def list_to_str(ConjunctiveSet, conjunctions: List, conjunct_to_str=lambda c: str(c)) -> str:
        """
        List should be a list of ConjunctiveSet
        Given [ConjunctiveSet, ConjunctiveSet], this will call ConjunctiveSet.display() using the `conjunct_to_str` function
        (which defaults to `str`) and join them with a pipe.

        To have elements of the ConjunctiveSet displayed differently, use a different `conjunct_to_str` function. 
        This functionality is used to differentiate between GPAD 1.2 and GPAD 2.0, where relations are written 
        differently per version.
        """
        return "|".join([conj.display(conjunct_to_str=conjunct_to_str) for conj in conjunctions])

    @classmethod
    def str_to_conjunctions(ConjunctiveSet, entity: str, conjunct_element_builder: Union[C, Error]=lambda el: Curie.from_str(el)) -> Union[List[C], Error]:
        """
        Takes a field that conforms to the pipe (|) and comma (,) separator type. The parsed version is a list of pipe separated values
        which are themselves a comma separated list.

        If the elements inside the comma separated list should not just be strings, but be converted into a value of a type, `conjunct_element_builder` can be provided which should take a string and return a parsed value or an instance of an Error type (defined above).

        If there is an error in producing the values of the conjunctions, then this function will return early with the error.

        This function will return a List of ConjunctiveSet
        """
        conjunctions = []
        for conj in filter(None, entity.split("|")):
            conjunct = []
            for el in filter(None, conj.split(",")):
                built = conjunct_element_builder(el)
                if isinstance(built, Error):
                    # Returning an Error instance
                    return built

                conjunct.append(built)

            conjunctions.append(ConjunctiveSet(conjunct))

        return conjunctions

    def is_error(self) -> bool:
        return False

@dataclass(unsafe_hash=True)
class Evidence:
    type: Curie # Curie of the ECO class
    has_supporting_reference: List[Curie]
    with_support_from: List[ConjunctiveSet]


relation_tuple = re.compile(r'([\w]+)\((\w+:[\w][\w\.:\-]*)\)')
curie_relation_tuple = re.compile(r"(.+)\((.+)\)")
@dataclass(unsafe_hash=True)
class ExtensionUnit:
    """
    An ExtensionUnit is a single element of the extensions field of GAF or GPAD. This consists of a relation and a term.

    Create an ExtensionUnit with `from_str` or `from_curie_str`. If there is an error in parsing then `Error` is returned.
    Results from these functions should be checked for `Error`.

    The string representation will depend on the format, and so the `display` method should be used. By default this
    will write the relation using the label with undercores (example: part_of) as defined in ontobio.rdfgen.relations.py.
    To write the relation as a CURIE (as in gpad 2.0), set parameter `use_rel_label` to `True`.
    """
    relation: Curie
    term: Curie

    @classmethod
    def from_str(ExtensionUnit, entity: str) -> Union:
        """
        Attempts to parse string entity as an ExtensionUnit
        If the `relation(term)` is not formatted correctly, an Error is returned.
        If the `relation` cannot be found in the `relations` dictionary then an error
        is also returned.
        """
        parsed = relation_tuple.findall(entity)
        if len(parsed) == 1:
            rel, term = parsed[0]
            rel_uri = relations.lookup_label(rel)
            if rel_uri is None:
                # print("Error because rel_uri isn't in the file: {}".format(rel))
                return Error(entity)

            term_curie = Curie.from_str(term)
            rel_curie = relations.obo_uri_to_curie(rel_uri)
            if isinstance(term_curie, Error):
                # print("Error because term is screwed up: {}".format(term))
                return Error("`{}`: {}".format(term, term_curie.info))
            return ExtensionUnit(rel_curie, term_curie)
        else:
            # print("Just couldn't even parse it at all: {}".format(entity))
            return Error(entity)

    @classmethod
    def from_curie_str(ExtensionUnit, entity: str) -> Union:
        """
        Attempts to parse string entity as an ExtensionUnit
        If the `relation(term)` is not formatted correctly, an Error is returned.
        `relation` is a Curie, and so is any errors in formatting are delegated to Curie.from_str()
        """
        parsed = curie_relation_tuple.findall(entity)
        if len(parsed) == 1:
            rel, term = parsed[0]

            rel_curie = Curie.from_str(rel)
            term_curie = Curie.from_str(term)
            if term_curie.is_error():
                return Error("`{}`: {}".format(term, term_curie.info))
            if rel_curie.is_error():
                return Error("`{}`: {}".format(rel, rel_curie.info))
            return ExtensionUnit(rel_curie, term_curie)

        else:
            return Error(entity)

    def __str__(self) -> str:
        """
        Display Curie(term)
        """
        return self.display()

    def display(self, use_rel_label=False):
        """
        Turns the ExtensionUnit into a string. By default this uses the ontobio.rdfgen.relations module to lookup the
        relation label. To use the CURIE instead, pass `use_rel_label=True`.
        """
        rel = str(self.relation)
        if use_rel_label:
            rel = self.__relation_to_label()

        return "{rel}({term})".format(rel=rel, term=self.term)

    def __relation_to_label(self) -> str:
        # Curie -> expand to URI -> reverse relation lookup Label
        return relations.lookup_uri(relations.curie_to_obo_uri(self.relation))

    def to_hash(self, use_label=False) -> dict:
        rel = self.__relation_to_label() if use_label else str(self.relation)
        return {
            "property": rel,
            "filler": str(self.term)
        }

@dataclass(repr=True, unsafe_hash=True)
class GoAssociation:
    """
    The internal model used by the parsers and qc Rules engine that all annotations are parsed into.

    If an annotation textual line cannot be parsed into a GoAssociation then it is not a well formed line.

    This class provides several methods to convert this GoAssociation into other representations, like GAF and GPAD
    of each version, as well as the old style dictionary Association that this class replaced (for compatibility if needed).

    Each parser has its own function or functions that converts an annotation line into a GoAssociation, and this is the first
    phase of parsing. In general, GoAssociations are only created by the parsers.
    """
    source_line: Optional[str]
    subject: Subject
    relation: Curie # This is the relation Curie
    object: Term
    negated: bool
    qualifiers: List[Curie]
    aspect: Optional[Aspect]
    interacting_taxon: Optional[Curie]
    evidence: Evidence
    subject_extensions: List[ExtensionUnit]
    object_extensions: List[ConjunctiveSet]
    provided_by: Provider
    date: Date
    properties: List[Tuple[str, str]]

    def to_gaf_2_1_tsv(self) -> List:
        """
        Converts the GoAssociation into a "TSV" columnar GAF 2.1 row as a list of strings.
        """
        gp_isoforms = "" if not self.subject_extensions else self.subject_extensions[0].term

        allowed_qualifiers = {"contributes_to", "colocalizes_with"}

        # Curie Object -> CURIE Str -> URI -> Label
        qual_labels = [relations.lookup_uri(curie_util.expand_uri(str(q), strict=False)) for q in self.qualifiers]
        if len(qual_labels) == 1 and qual_labels[0] not in allowed_qualifiers:
            logger.warning("Cannot write qualifier `{}` in GAF version 2.1 since only {} are allowed: skipping".format(self.qualifiers[0], ", ".join(allowed_qualifiers)))
            # If the qualifier is wrong, blank out the qualifiers
            qual_labels = []

        if self.negated:
            qual_labels.append("NOT")

        qualifier = "|".join(qual_labels)

        self.object.taxon.namespace = "taxon"
        taxon = str(self.object.taxon)
        if self.interacting_taxon:
            self.interacting_taxon.namespace = "taxon"
            taxon = "{taxon}|{interacting}".format(taxon=taxon, interacting=str(self.interacting_taxon))

        # For extensions, we provide the to string function on ConjunctElement that
        # calls its `display` method, with the flag to use labels instead of the CURIE.
        # This function is used to turn the whole column correctly into a string
        return [
            self.subject.id.namespace,
            self.subject.id.identity,
            self.subject.label,
            qualifier,
            str(self.object.id),
            "|".join([str(ref) for ref in self.evidence.has_supporting_reference]),
            ecomap.ecoclass_to_coderef(str(self.evidence.type))[0],
            ConjunctiveSet.list_to_str(self.evidence.with_support_from),
            self.aspect if self.aspect else "",
            self.subject.fullname_field(),
            "|".join(self.subject.synonyms),
            gp_type_label_to_curie(self.subject.type[0]),
            taxon,
            ymd_str(self.date, ""),
            self.provided_by,
            ConjunctiveSet.list_to_str(self.object_extensions,
                conjunct_to_str=lambda conj: conj.display(use_rel_label=True)),
            gp_isoforms
        ]

    def to_gaf_2_2_tsv(self) -> List:
        """
        Converts the GoAssociation into a "TSV" columnar GAF 2.2 row as a list of strings.
        """
        gp_isoforms = "" if not self.subject_extensions else self.subject_extensions[0].term

        qual_labels = [relations.lookup_uri(curie_util.expand_uri(str(q), strict=False)) for q in self.qualifiers]
        if self.negated:
            qual_labels.insert(0, "NOT")

        qualifier = "|".join(qual_labels)

        self.object.taxon.namespace = "taxon"
        taxon = str(self.object.taxon)
        if self.interacting_taxon:
            self.interacting_taxon.namespace = "taxon"
            taxon = "{taxon}|{interacting}".format(taxon=taxon, interacting=str(self.interacting_taxon))

        return [
            self.subject.id.namespace,
            self.subject.id.identity,
            self.subject.label,
            qualifier,
            str(self.object.id),
            "|".join([str(ref) for ref in self.evidence.has_supporting_reference]),
            ecomap.ecoclass_to_coderef(str(self.evidence.type))[0],
            ConjunctiveSet.list_to_str(self.evidence.with_support_from),
            self.aspect if self.aspect else "",
            self.subject.fullname_field(),
            "|".join(self.subject.synonyms),
            gp_type_label_to_curie(self.subject.type[0]),
            taxon,
            ymd_str(self.date, ""),
            self.provided_by,
            ConjunctiveSet.list_to_str(self.object_extensions,
                conjunct_to_str=lambda conj: conj.display(use_rel_label=True)),
            gp_isoforms
        ]

    def to_gpad_1_2_tsv(self) -> List:
        """
        Converts the GoAssociation into a "TSV" columnar GPAD 1.2 row as a list of strings.
        """
        # Curie Object -> CURIE Str -> URI -> Label
        qual_labels = [relations.lookup_uri(curie_util.expand_uri(str(q), strict=False)) for q in self.qualifiers]

        # Try qualifiers first since, if we are going from GAF -> GPAD and the GAF had a qualifier, that would be
        # more specific than the relation, which is calculated from the aspect/Go term.
        if qual_labels == []:
            # If there were no qualifiers, then we'll use the Relation. Gpad requires at least one qualifier (which is the relation)
            qual_labels.append(relations.lookup_uri(curie_util.expand_uri(str(self.relation), strict=False)))

        if self.negated:
            qual_labels = ["NOT"] + qual_labels

        qualifier = "|".join(qual_labels)

        props_list = ["{key}={value}".format(key=key, value=value) for key, value in self.properties]
        return [
            self.subject.id.namespace,
            self.subject.id.identity,
            qualifier,
            str(self.object.id),
            "|".join([str(ref) for ref in self.evidence.has_supporting_reference]),
            str(self.evidence.type),
            ConjunctiveSet.list_to_str(self.evidence.with_support_from),
            str(self.interacting_taxon) if self.interacting_taxon else "",
            ymd_str(self.date, ""),
            self.provided_by,
            ConjunctiveSet.list_to_str(self.object_extensions,
                conjunct_to_str=lambda conj: conj.display(use_rel_label=True)),
            "|".join(props_list)
        ]

    def to_gpad_2_0_tsv(self) -> List:
        """
        Converts the GoAssociation into a "TSV" columnar GAF 2.0 row as a list of strings.
        """

        props_list = ["{key}={value}".format(key=key, value=value) for key, value in self.properties]
        return [
            str(self.subject.id),
            "NOT" if self.negated else "",
            str(self.relation),
            str(self.object.id),
            "|".join([str(ref) for ref in self.evidence.has_supporting_reference]),
            str(self.evidence.type),
            ConjunctiveSet.list_to_str(self.evidence.with_support_from),
            str(self.interacting_taxon) if self.interacting_taxon else "",
            ymd_str(self.date, "-"),
            self.provided_by,
            ConjunctiveSet.list_to_str(self.object_extensions,
                conjunct_to_str=lambda conj: conj.display()),
            "|".join(props_list)
        ]

    def to_hash_assoc(self) -> dict:
        """
        Converts the GoAssociation into the old style dictionary association for backwards compatibility
        """
        subject = {
            "id": str(self.subject.id),
            "label": self.subject.label,
            "type": gp_type_label_to_curie(self.subject.type[0]),
            "fullname": self.subject.fullname[0],
            "synonyms": self.subject.synonyms,
            "taxon": {
                "id": str(self.subject.taxon)
            }
        }

        obj = {
            "id": str(self.object.id),
            "taxon": str(self.object.taxon)
        }

        subject_extensions = [{"property": str(subj.relation), "filler": str(subj.term)} for subj in self.subject_extensions]

        disjunctions = []
        for conjset in self.object_extensions:
            conjunctions = []
            for extension in conjset.elements:
                conjunctions.append(extension.to_hash(use_label=True))
            disjunctions.append({"intersection_of": conjunctions})

        object_extensions = {}
        if len(disjunctions) > 0:
            object_extensions["union_of"] = disjunctions

        withfrom_flat = []
        for withfrom in self.evidence.with_support_from:
            for curie in withfrom.elements:
                withfrom_flat.append(str(curie))

        evidence = {
            "type": ecomap.ecoclass_to_coderef(str(self.evidence.type))[0],
            "has_supporting_reference": [str(ref) for ref in self.evidence.has_supporting_reference],
            "with_support_from": withfrom_flat
        }

        return {
            "source_line": self.source_line,
            "subject": subject,
            "object": obj,
            "negated": self.negated,
            "qualifiers": [str(q) for q in self.qualifiers],
            "aspect": self.aspect,
            "relation": {
                "id": str(self.relation)
            },
            "interacting_taxon": self.interacting_taxon,
            "evidence": evidence,
            "provided_by": self.provided_by,
            "date": ymd_str(self.date, ""),
            "subject_extensions": subject_extensions,
            "object_extensions": object_extensions
        }

    def annotation_property_values(self, property_key: str) -> List:
        fetched_prop_values = [value for key, value in self.properties if key == property_key]
        return fetched_prop_values

@dataclass
class Header:
    souce_line: Optional[str]



"""Simple association model

The core class here is AssociationSet, a holder for a set of
associations between entities such as genes and ontology
classes. AssociationSets can also be throught of as subsuming
traditional 'gene sets'

The model is deliberately simple, and does not seek to represent
metadata about the association - it is assumed that this is handled
upstream. See the assoc_factory module for details - this allows the
client to create an association set based on various criteria such as
taxa of interest or evidence criteria.

"""
import logging
import scipy.stats # TODO - move
import scipy as sp # TODO - move
import pandas as pd

logger = logging.getLogger(__name__)


class UnknownSubjectException(Exception):
    pass

class AssociationSet():
    """An object that represents a collection of associations

    NOTE: the intention is that this class can be subclassed to provide
    either high-efficiency implementations, or implementations backed by services or external stores.
    The default implementation is in-memory.

    """

    def __init__(self, ontology=None, association_map=None, subject_label_map=None, meta=None):
        """
        NOTE: in general you do not need to call this yourself. See assoc_factory

        initializes an association set, which minimally consists of:

         - an ontology (e.g. GO, HP)
         - a map between subjects (e.g genes) and sets/lists of term IDs

        """
        self.ontology = ontology
        self.association_map = association_map
        self.subject_label_map = subject_label_map
        self.subject_to_inferred_map = {}
        self.meta = meta  # TODO
        self.associations_by_subj = None
        self.associations_by_subj_obj = None
        self.strict = False
        self.index()

        if self.association_map is None:
            self.association_map = {}

        logger.info("Created {}".format(self))

    def __str__(self):
        imap = self.subject_to_inferred_map
        return "AssocSet |S|={} |S->I|={}".format(len(imap.keys()), len(imap.items()))

    def index(self):
        """
        Creates indexes based on inferred terms.

        You do not need to call this yourself; called on initialization
        """
        self.subjects = list(self.association_map.keys())

        # ensure annotations unique
        for (subj,terms) in self.association_map.items():
            self.association_map[subj] = list(set(self.association_map[subj]))
            
        logger.info("Indexing {} items".format(len(self.subjects)))
        n = 0
        all_objs = set()
        for (subj,terms) in self.association_map.items():
            ancs = self.termset_ancestors(terms)
            all_objs.update(ancs)
            self.subject_to_inferred_map[subj] = ancs
            n = n+1
            if n<5:
                logger.debug(" Indexed: {} -> {}".format(subj, ancs))
            elif n == 6:
                logger.info("[TRUNCATING>5]....")
        self.objects = all_objs

    def inferred_types(self, subj):
        """
        Returns: set of reflexive inferred types for a subject.

        E.g. if a gene is directly associated with terms A and B, and these terms have ancestors C, D and E
        then the set returned will be {A,B,C,D,E}
        
        Arguments
        ---------

          subj - ID string

        Returns: set of class IDs

        """
        if subj in self.subject_to_inferred_map:
            return self.subject_to_inferred_map[subj]
        if self.strict:
            raise UnknownSubjectException(subj)
        else:
            return set([])
        
    def termset_ancestors(self, terms):
        """
        reflexive ancestors

        Arguments
        ---------

          terms - a set or list of class IDs

        Returns: set of class IDs
        """
        ancs = set()
        for term in terms:
            ancs = ancs.union(self.ontology.ancestors(term))
        return ancs.union(set(terms))

    def query_associations(self, subjects=None, infer_subjects=True, include_xrefs=True):
        """
        Query for a set of associations.

        Note: only a minimal association model is stored, so all results are returned as (subject_id,class_id) tuples

        Arguments:

         subjects: list

          list of subjects (e.g. genes, diseases) used to query associations. Any association to one of these subjects or
          a descendant of these subjects (assuming infer_subjects=True) are returned.

         infer_subjects: boolean (default true)

          See above

         include_xrefs: boolean (default true)

          If true, then expand inferred subject set to include all xrefs of those subjects.

        Example: if a high level disease node (e.g. DOID:14330 Parkinson disease) is specified, then the default behavior
        (infer_subjects=True, include_xrefs=True) and the ontology includes DO, results will include associations from 
        both descendant DOID classes, and all xrefs (e.g. OMIM)

        """
        if subjects is None:
            subjects = []
        mset = set()
        if infer_subjects:
            for subj in subjects:
                mset.update(self.ontology.descendants(subj))
        mset.update(set(subjects))
        if include_xrefs:
            xset = set()
            for m in mset:
                xrefs = self.ontology.xrefs(m, bidirectional=True)
                if xrefs is not None:
                    xset.update(xrefs)
            mset.update(xset)
        logger.debug("Matching subjects: {}".format(mset))
        mset = mset.intersection(self.subjects)
        logger.debug("Matching subjects with anns: {}".format(mset))
        amap = self.association_map
        results = []
        for m in mset:
            if m in amap:
                for t in amap[m]:
                    results.append( (m,t) )
        return results

    def annotations(self, subject_id):
        """
        Returns a list of classes used to describe a subject

        @Deprecated: use objects_for_subject
        """
        if subject_id in self.association_map:
            return self.association_map[subject_id]
        else:
            return []

    def objects_for_subject(self, subject_id):
        """
        Returns a list of classes used to describe a subject
        """
        if subject_id in self.association_map:
            return self.association_map[subject_id]
        else:
            return []
        
    def query(self, terms=None, negated_terms=None):
        """
        Basic boolean query, using inference.

        Arguments:

         - terms: list

             list of class ids. Returns the set of subjects that have at least one inferred annotation to each of the specified classes.

         - negated_terms: list

             list of class ids. Filters the set of subjects so that there are no inferred annotations to any of the specified classes
        """

        if terms is None:
            terms = []
        matches_all = 'owl:Thing' in terms
        if negated_terms is None:
            negated_terms = []
        termset = set(terms)
        negated_termset = set(negated_terms)
        matches = []
        n_terms = len(termset)
        for subj in self.subjects:
            if matches_all or len(termset.intersection(self.inferred_types(subj))) == n_terms:
                if len(negated_termset.intersection(self.inferred_types(subj))) == 0:
                    matches.append(subj)
        return matches

    def query_intersections(self, x_terms=None, y_terms=None, symmetric=False):
        """
        Query for intersections of terms in two lists

        Return a list of intersection result objects with keys:
         - x : term from x
         - y : term from y
         - c : count of intersection
         - j : jaccard score
        """
        if x_terms is None:
            x_terms = []
        if y_terms is None:
            y_terms = []
        xset = set(x_terms)
        yset = set(y_terms)
        zset = xset.union(yset)

        # first built map of gene->termClosure.
        # this could be calculated ahead of time for all g,
        # but this may be space-expensive. TODO: benchmark
        gmap={}
        for z in zset:
            gmap[z] = []
        for subj in self.subjects:
            ancs = self.inferred_types(subj)
            for a in ancs.intersection(zset):
                gmap[a].append(subj)
        for z in zset:
            gmap[z] = set(gmap[z])
        ilist = []
        for x in x_terms:
            for y in y_terms:
                if not symmetric or x<y:
                    shared = gmap[x].intersection(gmap[y])
                    union = gmap[x].union(gmap[y])
                    j = 0
                    if len(union)>0:
                        j = len(shared) / len(union)
                    ilist.append({'x':x,'y':y,'shared':shared, 'c':len(shared), 'j':j})
        return ilist

    @staticmethod
    def intersectionlist_to_matrix(ilist, xterms, yterms):
        """
        WILL BE DEPRECATED

        Replace with method to return pandas dataframe
        """
        z = [ [0] * len(xterms) for i1 in range(len(yterms)) ]
    
        xmap = {}
        xi = 0
        for x in xterms:
            xmap[x] = xi
            xi = xi+1
        
        ymap = {}
        yi = 0
        for y in yterms:
            ymap[y] = yi
            yi = yi+1
            
        for i in ilist:
            z[ymap[i['y']]][xmap[i['x']]] = i['j']
            
        logger.debug("Z={}".format(z))
        return (z,xterms,yterms)
    
    def as_dataframe(self, fillna=True, subjects=None):
        """
        Return association set as pandas DataFrame

        Each row is a subject (e.g. gene)
        Each column is the inferred class used to describe the subject
        """
        entries = []
        selected_subjects = self.subjects
        if subjects is not None:
            selected_subjects = subjects
            
        for s in selected_subjects:
            vmap = {}
            for c in self.inferred_types(s):
                vmap[c] = 1
            entries.append(vmap)
        logger.debug("Creating DataFrame")
        df = pd.DataFrame(entries, index=selected_subjects)
        if fillna:
            logger.debug("Performing fillna...")
            df = df.fillna(0)
        return df

    def label(self, id):
        """
        return label for a subject id

        Will make use of both the ontology and the association set
        """
        if self.ontology is not None:
            label = self.ontology.label(id)
            if label is not None:
                return label
        if self.subject_label_map is not None and id in self.subject_label_map:
            return self.subject_label_map[id]
        return None

    def subontology(self, minimal=False):
        """
        Generates a sub-ontology based on associations
        """
        return self.ontology.subontology(self.objects, minimal=minimal)

    def associations(self, subject, object=None):
        """
        Given a subject-object pair (e.g. gene id to ontology class id), return all association
        objects that match.

        """
        if object is None:
            if self.associations_by_subj is not None:
                return self.associations_by_subj[subject]
            else:
                return []
        else:
            if self.associations_by_subj_obj is not None:
                return self.associations_by_subj_obj[(subject,object)]
            else:
                return []

    # TODO: consider moving to other module
    def enrichment_test(self, subjects=None, background=None, hypotheses=None, threshold=0.05, labels=False, direction='greater'):
        """
        Performs term enrichment analysis. 

        Arguments
        ---------

        subjects: string list

            Sample set. Typically a gene ID list. These are assumed to have associations

        background: string list

            Background set. If not set, uses full set of known subject IDs in the association set

        threshold: float

            p values above this are filtered out

        labels: boolean

            if true, labels for enriched classes are included in result objects

        direction: 'greater', 'less' or 'two-sided'

            default is greater - i.e. enrichment test. Use 'less' for depletion test.

        """
        if subjects is None:
            subjects = []

        subjects=set(subjects)
        bg_count = {}
        sample_count = {}
        potential_hypotheses = set()
        sample_size = len(subjects)
        for s in subjects:
            potential_hypotheses.update(self.inferred_types(s))
        if hypotheses is None:
            hypotheses = potential_hypotheses
        else:
            hypotheses = potential_hypotheses.intersection(hypotheses)
        logger.info("Hypotheses: {}".format(hypotheses))
        
        # get background counts
        # TODO: consider doing this ahead of time
        if background is None:
            background = set(self.subjects)
        else:
            background = set(background)

        # ensure background includes all subjects
        background.update(subjects)
        
        bg_size = len(background)
        
        for c in hypotheses:
            bg_count[c] = 0
            sample_count[c] = 0
        for s in background:
            ancs = self.inferred_types(s)
            for a in ancs.intersection(hypotheses):
                bg_count[a] = bg_count[a]+1
        for s in subjects:
            for a in self.inferred_types(s):
                if a in hypotheses:
                    sample_count[a] = sample_count[a]+1

        hypotheses = [x for x in hypotheses if bg_count[x] > 1]
        logger.info("Filtered hypotheses: {}".format(hypotheses))
        num_hypotheses = len(hypotheses)
                
        results = []
        for cls in hypotheses:
            
            # https://en.wikipedia.org/wiki/Fisher's_exact_test
            #
            #              Cls  NotCls    RowTotal
            #              ---  ------    ---
            # study/sample [a,      b]    sample_size
            # rest of ref  [c,      d]    bg_size - sample_size
            #              ---     ---
            #              nCls  nNotCls

            a = sample_count[cls]
            b = sample_size - a
            c = bg_count[cls] - a
            d = (bg_size - bg_count[cls]) - b
            #logger.debug("ABCD="+str((cls,a,b,c,d,sample_size)))
            _, p_uncorrected = sp.stats.fisher_exact( [[a, b], [c, d]], direction)
            p = p_uncorrected * num_hypotheses
            if p>1.0:
                p=1.0
            #logger.debug("P={} uncorrected={}".format(p,p_uncorrected))
            if p<threshold:
                res = {'c':cls,'p':p,'p_uncorrected':p_uncorrected}
                if labels:
                    res['n'] = self.ontology.label(cls)
                results.append(res)
            
        results = sorted(results, key=lambda x:x['p'])
        return results
            
    def jaccard_similarity(self,s1,s2):
        """
        Calculate jaccard index of inferred associations of two subjects

        |ancs(s1) /\ ancs(s2)|
        ---
        |ancs(s1) \/ ancs(s2)|

        """
        a1 = self.inferred_types(s1)
        a2 = self.inferred_types(s2)
        num_union = len(a1.union(a2))
        if num_union == 0:
            return 0.0
        return len(a1.intersection(a2)) / num_union

    def similarity_matrix(self, x_subjects=None, y_subjects=None, symmetric=False):
        """
        Query for similarity matrix between groups of subjects

        Return a list of intersection result objects with keys:
         - x : term from x
         - y : term from y
         - c : count of intersection
         - j : jaccard score
        """
        if x_subjects is None:
            x_subjects = []
        if y_subjects is None:
            y_subjects = []

        xset = set(x_subjects)
        yset = set(y_subjects)
        zset = xset.union(yset)

        # first built map of gene->termClosure.
        # this could be calculated ahead of time for all g,
        # but this may be space-expensive. TODO: benchmark
        gmap={}
        for z in zset:
            gmap[z] = self.inferred_types(z)
        ilist = []
        for x in x_subjects:
            for y in y_subjects:
                if not symmetric or x<y:
                    shared = gmap[x].intersection(gmap[y])
                    union = gmap[x].union(gmap[y])
                    j = 0
                    if len(union)>0:
                        j = len(shared) / len(union)
                    ilist.append({'x':x,'y':y,'shared':shared, 'c':len(shared), 'j':j})
        return self.intersectionlist_to_matrix(ilist, x_subjects, y_subjects)
    

class NamedEntity():
    """
    E.g. a gene etc
    """

    def __init__(self, id, label=None, taxon=None):
        self.id=id
        self.label=label
        self.taxon=taxon

class AssociationSetMetadata():
    """
    Information about how an association set is derived
    """

    def __init__(self, id=None, taxon=None, evidence=None, subject_category=None, object_category=None):
        self.id=id
        self.taxon=taxon
        



"""

Quick similarity calculations

"""

from ontobio.golr.golr_associations import search_associations, GolrFields
import scipy.stats # TODO - move
import scipy as sp # TODO - move

M=GolrFields()  

def get_object_closure(subject, object_category=None, **kwargs):
    """
    Find all terms used to annotate subject plus ancestors
    """
    results = search_associations(subject=subject,
                                  object_category=object_category,
                                  select_fields=[],
                                  facet_fields=[M.OBJECT_CLOSURE],
                                  facet_limit=-1,
                                  rows=0,
                                  **kwargs)
    return set(results['facet_counts'][M.OBJECT_CLOSURE].keys())

def subject_pair_overlap(subject1, subject2, object_category=None, **kwargs):
    """
    Jaccard similarity
    """
    set1 = get_object_closure(subject1,
                              object_category=object_category,
                              **kwargs)
    set2 = get_object_closure(subject2,
                              object_category=object_category,
                              **kwargs)
    return len(set1.intersection(set2)), len(set1.union(set2))

def subject_pair_simj(subject1, subject2, **kwargs):
    """
    Jaccard similarity
    """
    i, u = subject_pair_overlap(subject1, subject2, **kwargs)
    if i==0:
        return 0.0
    return i / u



from ontobio.validation import rules
from ontobio.io import assocparser
from ontobio.io import gafparser
from ontobio import ontol_factory

import json

ontology = ontol_factory.OntologyFactory().create("tests/resources/goslim_generic.json")
gaf_line = "PomBase\tSPBC11C11.03\tndc80\t\tGO:0000942\tPMID:11553715\tIDA\t\tC\tNMS complex subunit Ndc80\tndc10|tid3\tprotein\ttaxon:4896\t20150122\tPomBase"

def test_create_base_parser():
    parser = rules.create_base_parser(rules.FormatType.RDF)
    assert parser == None

    parser = rules.create_base_parser(rules.FormatType.GAF)
    assert type(parser) == gafparser.GafParser
    assert parser.config == assocparser.AssocParserConfig()

def test_validate_input():
    parser = rules.create_base_parser(rules.FormatType.GAF)
    normal_gaf_line = rules.normalize_tsv_row(17, gaf_line)
    example = rules.RuleExample("gorule-0000027", rules.ExampleType.FAIL, normal_gaf_line, rules.FormatType.GAF, False)
    config = assocparser.AssocParserConfig(
        ontology=ontology
    )

    parsed = rules.validate_input(example, parser, config=config)
    assert parsed.output == normal_gaf_line

def test_example_success_with_fail_example():
    # We expect this example to fail gorule 27, so expected is False
    parser = rules.create_base_parser(rules.FormatType.GAF)
    normalized = rules.normalize_tsv_row(17, gaf_line)
    example = rules.RuleExample("gorule-0000027", rules.ExampleType.FAIL, normalized, rules.FormatType.GAF, False)
    parsed = rules.validate_input(example, parser, config=assocparser.AssocParserConfig(ontology=ontology))

    success = rules.example_success(example, parsed)
    assert success == True

    # We falsely expect rule 1 to pass, so this test is expected to be False
    parser = rules.create_base_parser(rules.FormatType.GAF)
    normalized = rules.normalize_tsv_row(17, gaf_line)
    example = rules.RuleExample("gorule-0000002", rules.ExampleType.FAIL, normalized, rules.FormatType.GAF, False)
    parsed = rules.validate_input(example, parser, config=assocparser.AssocParserConfig(ontology=ontology))

    success = rules.example_success(example, parsed)
    assert success == False

def test_validate_example():
    parser = rules.create_base_parser(rules.FormatType.GAF)
    normalized = rules.normalize_tsv_row(17, gaf_line)
    example = rules.RuleExample("gorule-0000027", rules.ExampleType.FAIL, normalized, rules.FormatType.GAF, False)

    expected = rules.ValidationResult(example, False, True, "Valid")

    config = assocparser.AssocParserConfig(ontology=ontology)
    assert rules.validate_example(example, config=config) == expected

def test_format_parse():

    assert rules.format_from_string("rdf") == rules.FormatType.RDF
    assert rules.format_from_string("gaf") == rules.FormatType.GAF
    assert rules.format_from_string("gpad") == rules.FormatType.GPAD
    assert rules.format_from_string("foo") == None

def test_example_json_converted():

    j = {
        "id": "GORULE:0000027",
        "examples": {
            "fail": [
                {
                    "comment": "an example",
                    "format": "gaf",
                    "input": gaf_line
                }
            ]
        }
    }

    # normalized = rules.normalize_tsv_row(17, gaf_line)
    example = rules.RuleExample("gorule-0000027", rules.ExampleType.FAIL, gaf_line, rules.FormatType.GAF, False)
    assert rules.RuleExample.example_from_json(j) == [example]



class GocamgenException(Exception):
    def __init__(self, message: str):
        self.message = message

    def __str__(self):
        return self.message


class ModelRdfWriteException(GocamgenException):
    pass


class ShexException(GocamgenException):
    pass


class GeneErrorSet:
    def __init__(self):
        self.errors = {}

    def add_error(self, gene, error):
        if gene not in self.errors:
            self.errors[gene] = []
        self.errors[gene].append(error)



import logging
from typing import List, Dict
from ontobio.ontol_factory import OntologyFactory
from ontobio.io.gpadparser import GpadParser
from ontobio.io.assocparser import SplitLine
from ontobio.model.association import GoAssociation, ymd_str
from ontobio.rdfgen.gocamgen import errors

logger = logging.getLogger(__name__)

GPAD_PARSER = GpadParser()
BINDING_ROOT = "GO:0005488"  # binding
IPI_ECO_CODE = "ECO:0000353"


class CollapsedAssociationSet:
    def __init__(self, associations: List[GoAssociation], ontology, gpi_entities):
        self.associations = associations
        self.collapsed_associations = []
        self.assoc_dict = {}
        self.go_ontology = ontology
        self.gpi_entities = gpi_entities
    #
    # def setup_ontologies(self):
    #     if self.go_ontology is None:
    #         self.go_ontology = OntologyFactory().create("go")

    # def collapse_annotations(self):
        # Here we shall decide the distinct assertion instances going into the model
        # This will reduce/eliminate need to SPARQL model graph
        # Group by:
        # 		1. ID
        # 		2. qualifiers (normalize order; any array gotta do this)
        # 		3. primary term
        # 		4. With/From (if primary term is BINDING_ROOT or descendant)
        # 		5. Extensions
        # 	Collapse multiple:
        # 		1. Reference
        # 		2. Evidence Code
        #       3. With/From (if primary term is not BINDING_ROOT or descendant)
        # 		4. Source line
        # 		5. Date
        # 		6. Assigned by
        # 		7. Properties
        # self.setup_ontologies()

        for a in self.associations:
            # Header
            header_with_froms, line_with_froms = self.get_with_froms(a)  # Handle pipe separation according to import requirements
            cas = []
            for wf in header_with_froms:
                ca = self.find_or_create_collapsed_association(a, wf)
                association_line = CollapsedAssociationLine(a, line_with_froms)
                ca.lines.append(association_line)
                cas.append(ca)

    def find_or_create_collapsed_association(self, association: GoAssociation, with_from):
        query_header = {
            'subject': {
                'id': association.subject.id
            },
            'qualifiers': sorted(association.qualifiers),
            'object': {
                'id': association.object.id
            },
            'object_extensions': get_annot_extensions(association),
            'negated': association.negated
        }
        if with_from and "header" in with_from:
            query_header['evidence'] = {'with_support_from': sorted(with_from['header'])}
        # for ca in self.collapsed_associations:
        #     if ca.header == query_header:
        #         return ca
        ca = self.find_by_go_association(association)
        if ca:
            return ca
        else:
            ca = CollapsedAssociation(association)
            self.collapsed_associations.append(ca)
            return ca

    def find_by_go_association(self, association: GoAssociation):
        for ca in self.collapsed_associations:
            if ca.header_data_matches(association, self.gpi_entities):
                return ca
            # if ca.header == query_header:
            #     return ca

    def __iter__(self):
        return iter(self.collapsed_associations)

    def get_with_froms(self, annot: GoAssociation) -> List:
        header_with_froms, line_with_froms = [], []

        # The craziest example:
        # MGI:1  GO:0005515  IPI  PR:1,PR:2|WEIRD:1,PR:3|MGI:2
        source_line = annot.source_line
        vals = source_line.split("\t")
        with_from_col = vals[6]
        # Parse into array (by "|") of arrays (by ",")
        with_from_ds = []  # The list of lists
        for piped_with_from in with_from_col.split("|"):
            # Will be bypassing ontobio ID validation? Let's try teaming up with ontobio functions!
            split_line = SplitLine(line=source_line, values=vals, taxon="")  # req'd for error reporting in ontobio?
            validated_comma_with_froms = GPAD_PARSER.validate_pipe_separated_ids(piped_with_from, split_line,
                                                                                 empty_allowed=True, extra_delims=",")
            with_from_ds.append(validated_comma_with_froms)

        # Now arrange these into "header" and "line" values
        eco_code = str(annot.evidence.type)
        term = str(annot.object.id)
        is_binding = eco_code == IPI_ECO_CODE and BINDING_ROOT in self.go_ontology.ancestors(term, reflexive=True)
        if is_binding:
            # Using GPI, check with_froms for taxon equivalency to subj_id
            if self.gpi_entities:
                subject_id = str(annot.subject.id)
                if subject_id in self.gpi_entities:
                    subject_entity = self.gpi_entities[subject_id]
                else:
                    error_message = "Annotation Object ID '{}' missing from provided GPI. Skipping annotation translation.".format(subject_id)
                    logger.warning(error_message)
                    # Throw Exception and except-skip in model builder
                    raise errors.GocamgenException(error_message)
                values_separated = []
                for wf in with_from_ds:
                    wf_separated = {
                        "header": [],
                        "line": []
                    }
                    for wf_id in wf:
                        wf_entity = self.gpi_entities.get(wf_id)
                        if wf_entity and wf_entity.get("taxon") == subject_entity["taxon"]:
                            wf_separated['header'].append(wf_id)
                        else:
                            wf_separated['line'].append(wf_id)
                    values_separated.append(wf_separated)
                return values_separated
            else:
                # Everything is defaulted to header if no GPI available
                return [{"header": wf} for wf in with_from_ds]
        else:
            # Everything is defaulted to line if not binding
            return [{"line": wf} for wf in with_from_ds]

    def header_with_froms(self, association: GoAssociation):
        # Returns with_froms that should be header-level
        pass

    def line_with_froms(self, association: GoAssociation):
        # Returns with_froms that should be line-level
        pass


class CollapsedAssociation:
    def __init__(self, association: GoAssociation):
        # TODO: Refactor out complex dictionary structure - reuse GoAssociation
        # header = {
        #     'subject': {
        #         'id': association.subject.id
        #     },
        #     'qualifiers': sorted(association.qualifiers),
        #     'object': {
        #         'id': association.object.id
        #     },
        #     'object_extensions': get_annot_extensions(association),
        #     'negated': association.negated
        # }
        # self.header = header
        self.subject = association.subject
        self.object = association.object
        self.negated = association.negated
        self.qualifiers = sorted(association.qualifiers, key=lambda q: str(q))
        self.with_froms = []
        self.object_extensions = association.object_extensions
        self.lines: List[CollapsedAssociationLine] = []

    def header_data_matches(self, association: GoAssociation, gpi_entities):
        # Needs access to gpi_entities, which is currently only on CollapsedAssociationSet
        # self.with_froms == header_with_froms(association) and
        if self.subject_id() == str(association.subject.id) and \
           self.object_id() == str(association.object.id) and \
           self.negated == association.negated and \
           self.qualifiers == sorted(association.qualifiers, key=lambda q: str(q)) and \
           self.object_extensions == association.object_extensions:
            return True
        return False

    def subject_id(self):
        # if "subject" in self.header and "id" in self.header["subject"]:
        #     return self.header["subject"]["id"]
        return str(self.subject.id)

    def object_id(self):
        # if "object" in self.header and "id" in self.header["object"]:
        #     return self.header["object"]["id"]
        return str(self.object.id)

    def annot_extensions(self):
        # if "object_extensions" in self.header:
        #     return self.header["object_extensions"]
        return self.object_extensions

    def with_from(self):
        # if "evidence" in self.header and "with_support_from" in self.header["evidence"]:
        #     return self.header["evidence"]["with_support_from"]
        return self.with_froms

    def __str__(self):
        # TODO: Reconstruct GPAD format or original line - could mean multiple lines for each evidence
        return "{} - {}".format(self.subject_id(), self.object_id())

    def __iter__(self):
        return iter(self.lines)


def dedupe_extensions(extensions):
    new_extensions = []
    for i in extensions:
        if i not in new_extensions:
            new_extensions.append(i)
    return new_extensions


class CollapsedAssociationLine:
    def __init__(self, assoc: GoAssociation, with_from=None):
        self.source_line = assoc.source_line
        self.references = [str(ref) for ref in sorted(assoc.evidence.has_supporting_reference, key=lambda x: str(x))]
        self.evidence_code = str(assoc.evidence.type)
        self.date = assoc.date
        self.assigned_by = assoc.provided_by
        self.with_from = with_from
        self.annotation_properties = assoc.properties
        # self.annotation_properties = {}
        # for prop, vals in assoc.properties.items():
        #     self.annotation_properties[str(prop)] = vals

        # if "annotation_properties" in assoc:
        #     self.annotation_properties = assoc["annotation_properties"]

    def as_dict(self):
        ds = {
            "source_line": self.source_line,
            "evidence": {
                "type": self.evidence_code,
                "has_supporting_reference": self.references
            },
            "date": self.date,
            "provided_by": self.assigned_by,
        }
        if self.annotation_properties:
            ds["annotation_properties"] = self.annotation_properties
        if self.with_from:
            ds["evidence"]["with_support_from"] = self.with_from
        return ds


def get_annot_extensions(annot: GoAssociation):
    return annot.object_extensions


def extract_properties_from_string(prop_col):
    props = prop_col.split("|")
    props_dict = {}
    for p in props:
        k, v = p.split("=")
        if k in props_dict:
            props_dict[k].append(v)
        else:
            props_dict[k] = [v]
    return props_dict


def extract_properties(annot):
    cols = annot["source_line"].rstrip().split("\t")
    if len(cols) >= 12:
        prop_col = cols[11]
        annot["annotation_properties"] = extract_properties_from_string(prop_col)
    return annot



import pytest
import datetime
import yaml
import json

from ontobio.model import association
from ontobio.model.association import Curie
from ontobio.io import qc
from ontobio.io import gaference
from ontobio.io import assocparser
from ontobio.io import gafparser
from ontobio.io import gpadparser
from ontobio import ontol, ontol_factory

import copy

ontology = ontol_factory.OntologyFactory().create("tests/resources/goslim_generic.json")

def make_annotation(db="blah",
                    db_id="blah12345",
                    db_obj_symb="blah",
                    qualifier="",
                    negated=False,
                    goid="GO:1234567",
                    references="BLAH:54321",
                    evidence="IDA",
                    withfrom="BLAH:12345",
                    aspect="C",
                    db_obj_name="",
                    db_obj_syn="",
                    db_obj_type="blah",
                    taxon="taxon:12345",
                    date="20200330",
                    assigned_by="blah",
                    extension="",
                    gene_form_id="",
                    properties="",
                    version=None,
                    from_gaf=True):

    if from_gaf:
        qual_parse = assocparser.Qualifier2_1()
        if version == "2.2":
            qual_parse = assocparser.Qualifier2_2()

        annotation = [db, db_id, db_obj_symb, qualifier, goid, references, evidence, withfrom, aspect, db_obj_name, db_obj_syn, db_obj_type, taxon, date, assigned_by, extension, gene_form_id]

        return gafparser.to_association(annotation, qualifier_parser=qual_parse)
    else:
        if version != "2.0":
            # Default to 1.2 if we're anything but 2.0
            annotation = [db, db_id, qualifier, goid, references, evidence, withfrom, taxon, date, assigned_by, extension, properties]
            return gpadparser.to_association(annotation, version="1.2")
        else:
            annotation = ["{}:{}".format(db, db_id), "NOT" if negated else "", qualifier, goid, references, evidence, withfrom, taxon, date, assigned_by, extension, properties]
            return gpadparser.to_association(annotation, version="2.0")

def all_rules_config(ontology=None) -> assocparser.AssocParserConfig:
    return assocparser.AssocParserConfig(
        ontology=ontology,
        rule_set=assocparser.RuleSet.ALL
    )

def test_qc_result():
    assert qc.result(True, qc.FailMode.HARD) == qc.ResultType.PASS
    assert qc.result(False, qc.FailMode.HARD) == qc.ResultType.ERROR
    assert qc.result(False, qc.FailMode.SOFT) == qc.ResultType.WARNING

def test_go_rule02():

    assoc = make_annotation(qualifier="NOT", goid="GO:0005515").associations[0]

    test_result = qc.GoRule02().test(assoc, all_rules_config())
    assert test_result.result_type == qc.ResultType.WARNING

    assoc.negated = False
    test_result = qc.GoRule02().test(assoc, all_rules_config())
    assert test_result.result_type == qc.ResultType.PASS

    assoc.negated = True
    assoc.object.id = Curie.from_str("GO:0003674")
    test_result = qc.GoRule02().test(assoc, all_rules_config())
    assert test_result.result_type == qc.ResultType.PASS

def test_go_rule_06():

    assoc = make_annotation(goid="GO:0005575", evidence="HEP", aspect="C").associations[0]

    test_result = qc.GoRule06().test(assoc, all_rules_config(ontology=ontology))
    assert test_result.result_type == qc.ResultType.ERROR

    assoc.aspect = "P"
    assoc.object.id = Curie.from_str("GO:0008150")
    test_result = qc.GoRule06().test(assoc, all_rules_config(ontology=ontology))
    assert test_result.result_type == qc.ResultType.PASS

    assoc.evidence.type = Curie.from_str("ECO:0000501")
    test_result = qc.GoRule06().test(assoc, all_rules_config(ontology=ontology))
    assert test_result.result_type == qc.ResultType.PASS

def test_go_rule_07():

    assoc = make_annotation(goid="GO:0003824", evidence="IPI").associations[0]

    test_result = qc.GoRule07().test(assoc, all_rules_config(ontology=ontology))
    assert test_result.result_type == qc.ResultType.WARNING

    assoc.object.id = Curie.from_str("GO:1234567")
    test_result = qc.GoRule07().test(assoc, all_rules_config(ontology=ontology))
    assert test_result.result_type == qc.ResultType.PASS

    assoc.object.id = Curie.from_str("GO:0003824")
    assoc.evidence.type = "ECO:0000501" # Not IPI
    test_result = qc.GoRule07().test(assoc, all_rules_config(ontology=ontology))
    assert test_result.result_type == qc.ResultType.PASS

def test_go_rule08():

    assoc = make_annotation(goid="GO:0006810", evidence="IEA").associations[0]

    test_result = qc.GoRule08().test(assoc, all_rules_config(ontology=ontology))
    assert test_result.result_type == qc.ResultType.WARNING

    assoc.evidence.type = Curie.from_str("ECO:0000305")
    test_result = qc.GoRule08().test(assoc, all_rules_config(ontology=ontology))
    assert test_result.result_type == qc.ResultType.WARNING

    assoc.object.id = Curie.from_str("GO:0007049")  # do not manually annotate
    # evidence is IC, non IEA
    test_result = qc.GoRule08().test(assoc, all_rules_config(ontology=ontology))
    assert test_result.result_type == qc.ResultType.WARNING

    assoc.evidence.type = Curie.from_str("ECO:0000501") #IEA
    # IEA, but on not manual, so should pass
    test_result = qc.GoRule08().test(assoc, all_rules_config(ontology=ontology))
    assert test_result.result_type == qc.ResultType.PASS

    assoc.object.id = Curie.from_str("GO:0034655")  # neither
    assoc.evidence.type = Curie.from_str("ECO:0000305")
    test_result = qc.GoRule08().test(assoc, all_rules_config(ontology=ontology))
    assert test_result.result_type == qc.ResultType.PASS

def test_go_rule11():

    assoc = make_annotation(goid="GO:0003674", evidence="ND").associations[0]
    test_result = qc.GoRule11().test(assoc, all_rules_config())
    assert test_result.result_type == qc.ResultType.PASS

    # Bad GO ID
    assoc = make_annotation(goid="GO:1234567", evidence="ND").associations[0]
    test_result = qc.GoRule11().test(assoc, all_rules_config())
    assert test_result.result_type == qc.ResultType.ERROR

    # Not ND and not Root
    assoc = make_annotation(goid="GO:1234567", evidence="IEA").associations[0]
    test_result = qc.GoRule11().test(assoc, all_rules_config())
    assert test_result.result_type == qc.ResultType.PASS

    # Root, but not ND
    assoc = make_annotation(goid="GO:0003674", evidence="IEA").associations[0]
    test_result = qc.GoRule11().test(assoc, all_rules_config())
    assert test_result.result_type == qc.ResultType.ERROR

def test_go_rules_13():

    a = ["PomBase", "SPBC11B10.09", "cdc2", "", "GO:0007275", "PMID:21873635", "IBA", "PANTHER:PTN000623979|TAIR:locus:2099478", "P", "Cyclin-dependent kinase 1", "UniProtKB:P04551|PTN000624043", "protein", "taxon:284812", "20170228", "GO_Central", "", ""]
    assoc = gafparser.to_association(a).associations[0]
    gaferences = gaference.load_gaferencer_inferences_from_file("tests/resources/test.inferences.json")
    test_result = qc.GoRule13().test(assoc, assocparser.AssocParserConfig(annotation_inferences=gaferences, rule_set=assocparser.RuleSet.ALL))
    assert test_result.result_type == qc.ResultType.ERROR

    a = ["PomBase", "SPBC11B10.09", "cdc2", "", "GO:0007275", "PMID:21873635", "EXP", "PANTHER:PTN000623979|TAIR:locus:2099478", "P", "Cyclin-dependent kinase 1", "UniProtKB:P04551|PTN000624043", "protein", "taxon:284812", "20170228", "GO_Central", "", ""]
    assoc = gafparser.to_association(a).associations[0]
    gaferences = gaference.load_gaferencer_inferences_from_file("tests/resources/test.inferences.json")
    test_result = qc.GoRule13().test(assoc, assocparser.AssocParserConfig(annotation_inferences=gaferences, rule_set=assocparser.RuleSet.ALL))
    assert test_result.result_type == qc.ResultType.WARNING

    a = ["PomBase", "SPBC11B10.09", "cdc2", "NOT", "GO:0007275", "PMID:21873635", "EXP", "PANTHER:PTN000623979|TAIR:locus:2099478", "P", "Cyclin-dependent kinase 1", "UniProtKB:P04551|PTN000624043", "protein", "taxon:284812", "20170228", "GO_Central", "", ""]
    assoc = gafparser.to_association(a).associations[0]
    gaferences = gaference.load_gaferencer_inferences_from_file("tests/resources/test.inferences.json")
    test_result = qc.GoRule13().test(assoc, assocparser.AssocParserConfig(annotation_inferences=gaferences, rule_set=assocparser.RuleSet.ALL))
    assert test_result.result_type == qc.ResultType.PASS

    a = ["AspGD", "ASPL0000059928", "AN0127", "", "GO:0032258", "AspGD_REF:ASPL0000000005", "IEA", "SGD:S000001917", "P", "", "AN0127|ANID_00127|ANIA_00127", "gene_product", "taxon:227321", "20200201", "AspGD", "", ""]
    assoc = gafparser.to_association(a).associations[0]
    gaferences = gaference.load_gaferencer_inferences_from_file("tests/resources/test.inferences.json")
    test_result = qc.GoRule13().test(assoc, assocparser.AssocParserConfig(annotation_inferences=gaferences, rule_set=assocparser.RuleSet.ALL))
    assert test_result.result_type == qc.ResultType.ERROR

def test_go_rules_15():

    assoc = make_annotation(goid="GO:0044419", taxon="taxon:123|taxon:456").associations[0]

    ontology = ontol_factory.OntologyFactory().create("tests/resources/go-interspecies-20210520.json")

    test_result = qc.GoRule15().test(assoc, all_rules_config(ontology=ontology))
    assert test_result.result_type == qc.ResultType.PASS

    assoc.object.id = Curie.from_str("GO:1234567")
    test_result = qc.GoRule15().test(assoc, all_rules_config(ontology=ontology))
    assert test_result.result_type == qc.ResultType.WARNING

    assoc.object.id = Curie.from_str("GO:0002812")
    test_result = qc.GoRule15().test(assoc, all_rules_config(ontology=ontology))
    assert test_result.result_type == qc.ResultType.PASS

    assoc.object.id = Curie.from_str("GO:0044215")
    assoc.object.id = Curie.from_str("NCBITaxon:123")
    assoc.interacting_taxon = None # This is the important part, no interacting taxon
    test_result = qc.GoRule15().test(assoc, all_rules_config(ontology=ontology))
    assert test_result.result_type == qc.ResultType.PASS

def test_go_rule_16():
    # No GO term w/ID
    assoc = make_annotation(evidence="IC", withfrom="BLAH:12345").associations[0]

    test_result = qc.GoRule16().test(assoc, all_rules_config())
    assert test_result.result_type == qc.ResultType.ERROR

    # withfrom has GO term
    assoc.evidence.with_support_from = association.ConjunctiveSet.str_to_conjunctions("GO:0023456")

    test_result = qc.GoRule16().test(assoc, all_rules_config())
    assert test_result.result_type == qc.ResultType.PASS

    # Pipe
    assoc.evidence.with_support_from = association.ConjunctiveSet.str_to_conjunctions("GO:0012345|BLAH:54321")

    test_result = qc.GoRule16().test(assoc, all_rules_config())
    assert test_result.result_type == qc.ResultType.PASS

    # Empty withfrom
    assoc.evidence.with_support_from = []

    test_result = qc.GoRule16().test(assoc, all_rules_config())
    assert test_result.result_type == qc.ResultType.ERROR

    # Not IC
    assoc.evidence.type = Curie.from_str("ECO:0000501")
    assoc.evidence.with_support_from = association.ConjunctiveSet.str_to_conjunctions("BLAH:5555555|FOO:999999")

    test_result = qc.GoRule16().test(assoc, all_rules_config())
    assert test_result.result_type == qc.ResultType.PASS


def test_go_rule_17():
    # IDA with anything in withfrom
    assoc = make_annotation(evidence="IDA", withfrom="BLAH:12345").associations[0]

    test_result = qc.GoRule17().test(assoc, all_rules_config())
    assert test_result.result_type == qc.ResultType.WARNING

    # Nothing in withfrom, passes
    assoc.evidence.with_support_from = []
    test_result = qc.GoRule17().test(assoc, all_rules_config())
    assert test_result.result_type == qc.ResultType.PASS

def test_go_rule_18():
    # IDA with nothing in withfrom
    a = ["blah"] * 15
    a[3] = ""
    a[5] = "PMID:12345"
    a[6] = "IPI"
    a[7] = ""
    a[8] = "P"
    a[12] = "taxon:123"
    a[13] = "20200303"
    assoc = make_annotation(evidence="IPI", withfrom="").associations[0]

    test_result = qc.GoRule18().test(assoc, all_rules_config())
    assert test_result.result_type == qc.ResultType.WARNING

    # Something in withfrom, passes
    assoc.evidence.with_support_from = association.ConjunctiveSet.str_to_conjunctions("BLAH:12345")
    test_result = qc.GoRule18().test(assoc, all_rules_config())
    assert test_result.result_type == qc.ResultType.PASS

def test_go_rule26():

    config = assocparser.AssocParserConfig(
        ontology=ontology,
        paint=True,
        rule_set=assocparser.RuleSet.ALL
    )
    assoc = make_annotation(goid="GO:1234567", evidence="IBA").associations[0]
    # Pass due to IBA in paint
    test_result = qc.GoRule26().test(assoc, config)
    assert test_result.result_type == qc.ResultType.PASS

    config = assocparser.AssocParserConfig(
        ontology=ontology,
        paint=False,
        rule_set=assocparser.RuleSet.ALL
    )
    assoc = make_annotation(goid="GO:1234567", evidence="IPI").associations[0]
    # Pass due to non IBA
    test_result = qc.GoRule26().test(assoc, config)
    assert test_result.result_type == qc.ResultType.PASS

    config = assocparser.AssocParserConfig(
        ontology=ontology,
        paint=False,
        rule_set=assocparser.RuleSet.ALL
    )
    assoc = make_annotation(goid="GO:1234567", evidence="IBA").associations[0]
    # Fail  due to non paint
    test_result = qc.GoRule26().test(assoc, config)
    assert test_result.result_type == qc.ResultType.ERROR

def test_go_rule28():

    config = all_rules_config(ontology=ontology)

    assoc = make_annotation(goid="GO:0005975", evidence="IDA", aspect="P").associations[0]
    test_result = qc.GoRule28().test(assoc, config)

    assert test_result.result_type == qc.ResultType.PASS
    assert test_result.result == assoc

    assoc = make_annotation(goid="GO:0005975", evidence="IDA", aspect="C").associations[0]
    test_result = qc.GoRule28().test(assoc, config)

    assert test_result.result_type == qc.ResultType.WARNING
    fixed_assoc = copy.deepcopy(assoc)
    fixed_assoc.aspect = "P"
    assert test_result.result == fixed_assoc
    assert test_result.message == "Found violation of: `Aspect can only be one of C, P, F` but was repaired"

def test_go_rule29():
    # Nov 11, 1990, more than a year old
    assoc = make_annotation(evidence="IEA", date="19901111").associations[0]

    ## Base test: old IEA.
    test_result = qc.GoRule29().test(assoc, all_rules_config())
    assert test_result.result_type == qc.ResultType.ERROR

    ## Pass if not IEA
    assoc.evidence.type = Curie.from_str("ECO:0000305")  # Not IEA
    test_result = qc.GoRule29().test(assoc, all_rules_config())
    assert test_result.result_type == qc.ResultType.PASS

    ## Pass if only a half year old.
    now = datetime.datetime.now()
    six_months_ago = now - datetime.timedelta(days=180)
    assoc.date = association.Date(six_months_ago.year, six_months_ago.month, six_months_ago.day, "")
    assoc.evidence.type = Curie.from_str("ECO:0000501")
    test_result = qc.GoRule29().test(assoc, all_rules_config())
    assert test_result.result_type == qc.ResultType.PASS

    ## Warning if a year and a half year old.
    eighteen_months_ago = now - datetime.timedelta(days=(30*18))
    assoc.date = association.Date(eighteen_months_ago.year, eighteen_months_ago.month, eighteen_months_ago.day, "")
    test_result = qc.GoRule29().test(assoc, all_rules_config())
    assert test_result.result_type == qc.ResultType.WARNING

    ## Confirm the test can parse a YYYY-MM-DD date format from GPAD 2.0
    gpad_2_0_vals = assoc.to_gpad_2_0_tsv()  # Cheat to shortcut DB and DB_Object_ID concatenation
    gpad_2_0_vals[5] = "ECO:0000501"
    gpad_2_0_vals[8] = "1990-11-11"
    assoc = gpadparser.to_association(gpad_2_0_vals, version="2.0").associations[0]
    test_result = qc.GoRule29().test(assoc, all_rules_config())
    assert test_result.result_type == qc.ResultType.ERROR

def test_gorule30():
    assoc = make_annotation(references="GO_REF:0000033").associations[0]

    config = assocparser.AssocParserConfig(
        goref_metadata={
            "goref-0000033": {
                "authors": "Pascale Gaudet, Michael Livstone, Paul Thomas, The Reference Genome Project",
                "id": "GO_REF:0000033",
                "is_obsolete": True
            }
        },
        rule_set=assocparser.RuleSet.ALL
    )

    test_result = qc.GoRule30().test(assoc, config)
    assert test_result.result_type == qc.ResultType.WARNING

    assoc.evidence.has_supporting_reference = [Curie.from_str("GO_PAINT:0000000")]
    test_result = qc.GoRule30().test(assoc, config)
    assert test_result.result_type == qc.ResultType.WARNING

    assoc.evidence.has_supporting_reference = [Curie.from_str("FOO:123"), Curie.from_str("GO_REF:0000033")]
    test_result = qc.GoRule30().test(assoc, config)
    assert test_result.result_type == qc.ResultType.WARNING

    assoc.evidence.has_supporting_reference = [Curie.from_str("FOO:123")]
    test_result = qc.GoRule30().test(assoc, config)
    assert test_result.result_type == qc.ResultType.PASS

def test_gorule37():
    assoc = make_annotation(evidence="IBA", references="PMID:21873635", assigned_by="GO_Central").associations[0]

    test_result = qc.GoRule37().test(assoc, all_rules_config())
    assert test_result.result_type == qc.ResultType.PASS

    assoc.evidence.type = Curie.from_str("ECO:0000305") # Rule doesn't apply, not IBA
    test_result = qc.GoRule37().test(assoc, all_rules_config())
    assert test_result.result_type == qc.ResultType.PASS

    assoc.evidence.type = Curie.from_str("ECO:0000318")
    assoc.evidence.has_supporting_reference = [Curie.from_str("GO_REF:123")]  # IBA, but wrong ref
    test_result = qc.GoRule37().test(assoc, all_rules_config())
    assert test_result.result_type == qc.ResultType.ERROR

    assoc.evidence.has_supporting_reference = [Curie.from_str("PMID:21873635")]
    assoc.provided_by = "Pascale"  # IBA, but wrong assigned_by
    test_result = qc.GoRule37().test(assoc, all_rules_config())
    assert test_result.result_type == qc.ResultType.ERROR

def test_gorule39():
    assoc = make_annotation(db="ComplexPortal", goid="GO:0032991").associations[0]

    test_result = qc.GoRule39().test(assoc, all_rules_config())
    assert test_result.result_type == qc.ResultType.ERROR

    assoc.subject.id = association.Curie("FB", "1234")
    test_result = qc.GoRule39().test(assoc, all_rules_config())
    assert test_result.result_type == qc.ResultType.PASS

    assoc.subject.id = association.Curie("ComplexPortal", "12345")
    assoc.object.id = association.Curie("GO", "0000023")
    test_result = qc.GoRule39().test(assoc, all_rules_config())
    assert test_result.result_type == qc.ResultType.PASS

def test_gorule42():
    assoc = make_annotation(qualifier="NOT", evidence="IKR").associations[0]

    test_result = qc.GoRule42().test(assoc, all_rules_config())
    assert test_result.result_type == qc.ResultType.PASS

    assoc.evidence.type = "ECO:0000305" # Not IKR so this rule is fine
    test_result = qc.GoRule42().test(assoc, all_rules_config())
    assert test_result.result_type == qc.ResultType.PASS

    assoc.evidence.type = "ECO:0000320"
    assoc.negated = False  # Not negated, so wrong
    test_result = qc.GoRule42().test(assoc, all_rules_config())
    assert test_result.result_type == qc.ResultType.ERROR

def test_gorule43():
    assoc = make_annotation(references="GO_REF:0000024", evidence="ISO").associations[0]

    config = assocparser.AssocParserConfig(
        goref_metadata={
            "goref-0000024": {
                "authors": "Pascale Gaudet, Michael Livstone, Paul Thomas, The Reference Genome Project",
                "id": "GO_REF:0000024",
                "evidence_codes": ["ECO:0000266"]
            }
        },
        rule_set=assocparser.RuleSet.ALL
    )

    test_result = qc.GoRule43().test(assoc, config)
    assert test_result.result_type == qc.ResultType.PASS

    assoc.evidence.type = Curie.from_str("ECO:0000501")
    test_result = qc.GoRule43().test(assoc, config)
    assert test_result.result_type == qc.ResultType.WARNING

    assoc.evidence.type = Curie.from_str("ECO:0000266")
    assoc.evidence.has_supporting_reference = [Curie.from_str("FOO:123")]
    test_result = qc.GoRule43().test(assoc, config)
    assert test_result.result_type == qc.ResultType.PASS

def test_gorule46():
    config = all_rules_config(ontology=ontology)
    # Self-binding, yes
    assoc = make_annotation(db="PomBase", db_id="SPAC25B8.17", goid="GO:0051260", withfrom="PomBase:SPAC25B8.17").associations[0]

    test_result = qc.GoRule46().test(assoc, config)
    assert test_result.result_type == qc.ResultType.PASS

    assoc.evidence.with_support_from = association.ConjunctiveSet.str_to_conjunctions("PomBase:BLAH123")
    test_result = qc.GoRule46().test(assoc, config)
    assert test_result.result_type == qc.ResultType.WARNING

    assoc.evidence.with_support_from = association.ConjunctiveSet.str_to_conjunctions("PomBase:SPAC25B8.17|PomBase:BLAH123")
    test_result = qc.GoRule46().test(assoc, config)
    assert test_result.result_type == qc.ResultType.PASS

    assoc.object.id = Curie.from_str("GO:0000123")
    # Not in a self-binding mode
    test_result = qc.GoRule46().test(assoc, config)
    assert test_result.result_type == qc.ResultType.PASS

    # Test no ontology should just pass
    assoc.object.id = Curie.from_str("GO:0051260")
    test_result = qc.GoRule46().test(assoc, all_rules_config())
    assert test_result.result_type == qc.ResultType.PASS

def test_gorule50():
    a = "PMID:21873635"
    assoc = make_annotation(db="HELLO", db_id="123", evidence="ISS", withfrom="HELLO:123").associations[0]

    test_result = qc.GoRule50().test(assoc, all_rules_config())
    assert test_result.result_type == qc.ResultType.WARNING

    assoc.subject.id = Curie.from_str("BYE:567")
    test_result = qc.GoRule50().test(assoc, all_rules_config())
    assert test_result.result_type == qc.ResultType.PASS

    # Not ISS, so fine to have repeated columns
    assoc.subject.id = Curie.from_str("HELLO:123")
    assoc.evidence.type = Curie.from_str("ECO:0000501")
    test_result = qc.GoRule50().test(assoc, all_rules_config())
    assert test_result.result_type == qc.ResultType.PASS

def test_gorule55():
    a = ["blah"] * 15
    a[0] = "HELLO"
    a[1] = "123"
    a[3] = ""
    a[5] = "GO:0012345|PMID:1234567"
    a[6] = "ISS"
    a[7] = "HELLO:123"
    a[8] = "P"
    a[12] = "taxon:12345"
    a[13] = "20200303"
    assoc = make_annotation(db="HELLO", db_id="123", references="GO:0012345|PMID:1234567", evidence="ISS", withfrom="HELLO:123").associations[0]

    test_result = qc.GoRule55().test(assoc, all_rules_config())
    assert test_result.result_type == qc.ResultType.PASS

    assoc.evidence.has_supporting_reference = [association.Curie.from_str("GO:0001234"), association.Curie.from_str("GO:123456")]
    test_result = qc.GoRule55().test(assoc, all_rules_config())
    assert test_result.result_type == qc.ResultType.WARNING


def test_gorule57():
    assoc = make_annotation(db="HELLO", db_id="123", qualifier="contributes_to", goid="GO:0003674", evidence="ECO:0000501", taxon="taxon:2", from_gaf=False).associations[0]
    # Look at evidence_code, reference, annotation_properties
    config = assocparser.AssocParserConfig(
        group_metadata={
            "id": "mgi",
            "label": "Mouse Genome Informatics",
            "filter_out": {
                "evidence": ["ECO:0000501"],
                "evidence_reference": [
                    {
                        "evidence": "ECO:0000320",
                        "reference": "PMID:21873635"
                    }
                ],
                "annotation_properties": ["noctua-model-id"]
            }
        },
        rule_set=assocparser.RuleSet.ALL
    )
    test_result = qc.GoRule57().test(assoc, config)
    assert test_result.result_type == qc.ResultType.ERROR

    assoc.evidence.type = Curie.from_str("ECO:0000320")
    assoc.evidence.has_supporting_reference = [Curie.from_str("PMID:21873635")]
    test_result = qc.GoRule57().test(assoc, config)
    assert test_result.result_type == qc.ResultType.ERROR

    assoc.evidence.type = Curie.from_str("ECO:some_garbage")
    assoc.evidence.has_supporting_reference = [Curie.from_str("PMID:some_garbage")]
    assoc.properties = {"noctua-model-id": "some_garbage"}
    test_result = qc.GoRule57().test(assoc, config)
    assert test_result.result_type == qc.ResultType.ERROR

    assoc.properties = {}
    test_result = qc.GoRule57().test(assoc, config)
    assert test_result.result_type == qc.ResultType.PASS

def test_gorule58():

    with open("tests/resources/extensions-constraints.yaml") as exs_cons:
        config = assocparser.AssocParserConfig(ontology=ontology, extensions_constraints=yaml.load(exs_cons, Loader=yaml.FullLoader), rule_set=assocparser.RuleSet.ALL)

    assoc = make_annotation(db="HELLO", db_id="123", goid="GO:0003674", evidence="IBA", references="PMID:21873635", aspect="P", assigned_by="GO_Central", extension="has_input(GO:0003674),occurs_in(CL:123456)").associations[0]

    test_result = qc.GoRule58().test(assoc, config)
    assert test_result.result_type == qc.ResultType.PASS

    # Fails because `not_relation` is not an allowed relation
    new_ext = "coincident_with(GO:987),occurs_in(CL:1234567)|occurs_in(CL:12345)"
    assoc = make_annotation(db="HELLO", db_id="123", goid="GO:0003674", evidence="IBA", references="PMID:21873635", aspect="P", assigned_by="GO_Central", extension=new_ext).associations[0]
    test_result = qc.GoRule58().test(assoc, config)

    repaired_ext = "occurs_in(CL:12345)"
    expected_repair = make_annotation(db="HELLO", db_id="123", goid="GO:0003674", evidence="IBA", references="PMID:21873635", aspect="P", assigned_by="GO_Central", extension=repaired_ext).associations[0]
    expected_repair.source_line = assoc.source_line # Make sure original line is the same as test case
    assert test_result.result_type == qc.ResultType.WARNING
    assert test_result.result == expected_repair

    # Fails because `FOO` is not a real namespace for any `has_input` constraint
    new_ext = "has_input(FOO:1234566),occurs_in(CL:1234567)"
    assoc = make_annotation(db="HELLO", db_id="123", goid="GO:0003674", evidence="IBA", references="PMID:21873635", aspect="P", assigned_by="GO_Central", extension=new_ext).associations[0]
    test_result = qc.GoRule58().test(assoc, config)
    assert test_result.result_type == qc.ResultType.WARNING

    # Fails because this GO term is not in the constraint term children
    new_term = "GO:0005575"
    new_ext = "occurs_in(EMAPA:123),has_input(CL:1234567)"
    assoc = make_annotation(db="HELLO", db_id="123", goid=new_term, evidence="IBA", references="PMID:21873635", aspect="P", assigned_by="GO_Central", extension=new_ext).associations[0]
    test_result = qc.GoRule58().test(assoc, config)
    assert test_result.result_type == qc.ResultType.WARNING

    # Fails because of a cardinality check
    new_term = "GO:0003674"
    new_ext = "occurs_in(EMAPA:123),occurs_in(EMAPA:1987654)"
    assoc = make_annotation(db="HELLO", db_id="123", goid=new_term, evidence="IBA", references="PMID:21873635", aspect="P", assigned_by="GO_Central", extension=new_ext).associations[0]
    test_result = qc.GoRule58().test(assoc, config)
    assert test_result.result_type == qc.ResultType.WARNING

def test_gorule61():
    config = all_rules_config(ontology=ontology)
    assoc = make_annotation(goid="GO:0005554", qualifier="enables", evidence="ECO:0000320", from_gaf=False, version="1.2")
    assert assoc.report.reporter.messages.get("gorule-0000001", []) == []
    test_result = qc.GoRule61().test(assoc.associations[0], config)
    assert test_result.result_type == qc.ResultType.PASS

    # Using `contributes_to`, but should be repaired to RO:0002327 enables
    assoc = make_annotation(goid="GO:0005554", qualifier="contributes_to", evidence="ECO:0000320", from_gaf=False, version="1.2")
    test_result = qc.GoRule61().test(assoc.associations[0], config)
    assert test_result.result.relation == association.Curie("RO", "0002327")
    assert test_result.result_type == qc.ResultType.WARNING

    # BP term, qualifier inside allowed BP set
    assoc = make_annotation(goid="GO:0016192", qualifier="acts_upstream_of_or_within", evidence="ECO:0000320", from_gaf=False, version="1.2")
    test_result = qc.GoRule61().test(assoc.associations[0], config)
    assert test_result.result_type == qc.ResultType.PASS

    # BP term, unallowed relation, Repair
    assoc = make_annotation(goid="GO:0016192", qualifier="enables", evidence="ECO:0000320", from_gaf=False, version="1.2")
    test_result = qc.GoRule61().test(assoc.associations[0], config)
    assert test_result.result_type == qc.ResultType.WARNING
    assert test_result.result.relation == association.Curie("RO", "0002264")

    # CC complex term, unallowed relation, unrepairable, causes error
    assoc = make_annotation(goid="GO:0032991", qualifier="enables", evidence="ECO:0000320", from_gaf=False, version="1.2")
    test_result = qc.GoRule61().test(assoc.associations[0], config)
    assert test_result.result_type == qc.ResultType.ERROR

    # CC root repairs to is_active_in
    assoc = make_annotation(goid="GO:0005575", qualifier="located_in", evidence="ND", from_gaf=True, version="2.2")
    test_result = qc.GoRule61().test(assoc.associations[0], config)
    assert test_result.result_type == qc.ResultType.WARNING
    # Active in, rather than located_in
    assert test_result.result.relation == association.Curie(namespace="RO", identity="0002432")

    # protein complex + repairable relation repairs to part_of
    assoc = make_annotation(goid="GO:0032991", qualifier="is_active_in", evidence="ECO:0000320", from_gaf=False, version="1.2")
    test_result = qc.GoRule61().test(assoc.associations[0], config)
    assert test_result.result_type == qc.ResultType.WARNING
    assert test_result.result.relation == association.Curie(namespace="BFO", identity="0000050")


def test_all_rules():
    # pass
    config = all_rules_config(ontology=ontology)
    a = ["blah"] * 15
    a[3] = ""
    a[4] = "GO:0006397"
    a[5] = "PMID:21873635"
    a[6] = "ISS"
    a[7] = "PomBase:SPAC25B8.17"
    a[8] = "P"
    a[12] = "taxon:123"
    a[13] = "20180330"
    assoc = gafparser.to_association(a).associations[0]

    test_results = qc.test_go_rules(assoc, config).all_results
    assert len(test_results.keys()) == 24
    assert test_results[qc.GoRules.GoRule26.value].result_type == qc.ResultType.PASS
    assert test_results[qc.GoRules.GoRule29.value].result_type == qc.ResultType.PASS


if __name__ == "__main__":
    pytest.main(args=["tests/test_qc.py"])



"""
Classes for exporting entities.

So far only one implementation
"""
import re

def stringify(s):
    if s is None:
        return ""
    elif isinstance(s,list):
        return "|".join(s)
    else:
        return s

external_taxon = re.compile("taxon:([0-9]+)")
internal_taxon = re.compile("NCBITaxon:([0-9]+)")

def normalize_taxon(taxon):
    global internal_taxon
    global external_taxon

    if external_taxon.match(taxon):
        # If we match here, then the internal view already exists and we're good
        return internal_taxon

    match = internal_taxon.match(taxon)
    if match:
        taxon_id = match.group(1)
        return "taxon:{num}".format(num=taxon_id)

    return taxon


class EntityWriter():
    """
    Abstract superclass of all association writer objects (Gpad, GAF)
    """

    # TODO: add to superclass
    def _split_prefix(self, ref):
        id = ref['id']
        [prefix, local_id] = id.split(':', maxsplit=1)
        return prefix, local_id

    # TODO: add to superclass
    def _write_row(self, vals):
        vals = [stringify(v) for v in vals]
        line = "\t".join(vals)
        self.file.write(line + "\n")

    # TODO: add to superclass
    def write_entity(self, e):
        """
        Write a single entity
        """
        pass  ## Implemented in subclasses

    def write(self, entities, meta=None):
        """
        Write a complete set of entities to a file

        Arguments
        ---------
        entities: list[dict]
            A list of entity dict objects
        meta: Meta
            metadata about association set (not yet implemented)

        """
        for e in entities:
            self.write_entity(e)

class GpiWriter(EntityWriter):
    """
    Writes entities in GPI format

    Takes an entity dictionary:
    {
        'id': id, (String)
        'label': db_object_symbol, (String)
        'full_name': db_object_name, (String)
        'synonyms': synonyms, (List[str])
        'type': db_object_type, (String)
        'parents': parents, (List[Str])
        'xrefs': xref_ids, (List[Str])
        'taxon': {
            'id': self._taxon_id(taxon) (String)
        }
    }
    """
    def __init__(self, file=None):
        self.file = file
        if self.file:
            self.file.write("!gpi-version: 1.2\n")

    def write_entity(self, entity):
        """
        Write a single entity to a line in the output file
        """
        db, db_object_id = self._split_prefix(entity)
        taxon = normalize_taxon(entity["taxon"]["id"])

        vals = [
            db,
            db_object_id,
            entity.get('label'),
            entity.get('full_name'),
            entity.get('synonyms'),
            entity.get('type'),
            taxon,
            entity.get('parents'),
            entity.get('xrefs'),
            entity.get('properties')
        ]

        self._write_row(vals)



"""
Mapping between obograph-JSON format and networkx
"""

from ontobio.ontol import LogicalDefinition, PropertyChainAxiom
from ontobio.vocabulary.relations import map_legacy_pred
from ontobio.util.scigraph_util import get_curie_map
from ontobio.golr.golr_associations import search_associations

import json
import networkx
import logging
from prefixcommons.curie_util import expand_uri, contract_uri
from diskcache import Cache
import tempfile


cache = Cache(tempfile.gettempdir())


logger = logging.getLogger(__name__)


class OboJsonMapper(object):
    def __init__(self,
                 digraph=None,
                 context=None):
        self.digraph = digraph
        self.context = context if context is not None else {}

    def add_obograph_digraph(
            self,
            og,
            node_type=None,
            predicates=None,
            xref_graph=None,
            logical_definitions=None,
            property_chain_axioms=None,
            parse_meta=True,
            reverse_edges=True,
            **args):
        """
        Converts a single obograph to Digraph edges and adds to an existing networkx DiGraph
        """
        digraph = self.digraph
        logger.info("NODES: {}".format(len(og['nodes'])))

        # if client passes an xref_graph we must parse metadata
        if xref_graph is not None:
            parse_meta = True

        for node in og['nodes']:
            is_obsolete = 'is_obsolete' in node and node['is_obsolete'] == 'true'
            if is_obsolete:
                continue
            if node_type is not None and ('type' not in node or node['type'] != node_type):
                continue
            id = self.contract_uri(node['id'])
            digraph.add_node(id, **node)
            if 'lbl' in node:
                digraph.nodes[id]['label'] = node['lbl']
            if parse_meta and 'meta' in node:
                if node['meta'] is None:
                    node['meta'] = {}
                meta = self.transform_meta(node['meta'])
                if xref_graph is not None and 'xrefs' in meta:
                    for x in meta['xrefs']:
                        xref_graph.add_edge(self.contract_uri(x['val']), id, source=id)
        logger.info("EDGES: {}".format(len(og['edges'])))
        for edge in og['edges']:
            sub = self.contract_uri(edge['sub'])
            obj = self.contract_uri(edge['obj'])
            pred = self.contract_uri(edge['pred'])
            pred = map_legacy_pred(pred)
            if pred == 'is_a':
                pred = 'subClassOf'
            if predicates is None or pred in predicates:
                meta = edge['meta'] if 'meta' in edge else {}
                if reverse_edges:
                    digraph.add_edge(obj, sub, pred=pred, **meta)
                else:
                    digraph.add_edge(sub, obj, pred=pred, **meta)

        if 'equivalentNodesSets' in og:
            nslist = og['equivalentNodesSets']
            logger.info("CLIQUES: {}".format(len(nslist)))
            for ns in nslist:
                equivNodeIds = ns['nodeIds']
                for i in ns['nodeIds']:
                    ix = self.contract_uri(i)
                    for j in ns['nodeIds']:
                        if i != j:
                            jx = self.contract_uri(j)
                            digraph.add_edge(ix, jx, pred='equivalentTo')
        if logical_definitions is not None and 'logicalDefinitionAxioms' in og:
            for a in og['logicalDefinitionAxioms']:
                ld = LogicalDefinition(self.contract_uri(a['definedClassId']),
                                       [self.contract_uri(x) for x in a['genusIds']],
                                       [(self.contract_uri(x['propertyId']),
                                         self.contract_uri(x['fillerId'])) for x in a['restrictions'] if x is not None])
                logical_definitions.append(ld)
        if property_chain_axioms is not None and 'propertyChainAxioms' in og:
            for a in og['propertyChainAxioms']:
                pca = PropertyChainAxiom(predicate_id=self.contract_uri(a['predicateId']),
                                         chain_predicate_ids=[self.contract_uri(x) for x in a['chainPredicateIds']])
                property_chain_axioms.append(pca)

    def transform_meta(self, meta):
        if 'basicPropertyValues' in meta:
            for x in meta['basicPropertyValues']:
                x['pred'] = self.contract_uri(x['pred'])
                x['val'] = self.contract_uri(x['val'])
        return meta

    def contract_uri(self, uri):
        if len(self.context.keys()) > 0:
            curies = contract_uri(uri, cmaps=[self.context])
            if len(curies) > 0:
                return sorted(curies, key=len)[0] # sort by length

        curies = sorted(contract_uri(uri), key=len) # Sort by length
        if len(curies) > 0:
            return curies[0]
        else:
            return uri


def convert_json_file(obographfile, **args):
    """
    Return a networkx MultiDiGraph of the ontologies
    serialized as a json string

    """
    file = open(obographfile, 'r')
    jsonstr = file.read()
    file.close()
    return convert_json_object(json.loads(jsonstr), **args)


def convert_json_object(obographdoc, reverse_edges=True, **args):
    """
    Return a networkx MultiDiGraph of the ontologies
    serialized as a json object

    """
    digraph = networkx.MultiDiGraph()
    xref_graph = networkx.MultiGraph()
    logical_definitions = []
    property_chain_axioms = []
    context = obographdoc.get('@context',{})
    logger.info("CONTEXT: {}".format(context))
    mapper = OboJsonMapper(digraph=digraph, context=context)
    ogs = obographdoc['graphs']
    base_og = ogs[0]
    for og in ogs:
        # TODO: refactor this
        mapper.add_obograph_digraph(og, xref_graph=xref_graph,
                                    logical_definitions=logical_definitions,
                                    property_chain_axioms=property_chain_axioms,
                                    reverse_edges=reverse_edges,
                                    **args)

    return {
        'id': base_og.get('id'),
        'meta': base_og.get('meta'),
        'graph': mapper.digraph,
        'xref_graph': xref_graph,
        'graphdoc': obographdoc,
        'logical_definitions': logical_definitions,
        'property_chain_axioms': property_chain_axioms
        }


def _get_association_nodes(digraph, sub, predicate, obj):
    """
    Given a subject, predicate, object, retrieve the OBAN association
    node from the evidence graph
    """
    subject_assocs = set()
    object_assocs = set()
    predicate_assocs = set()
    union_assoc = set()

    # A fast way of checking if the triple is reified is seeing if the
    # predicate exists as a node
    if predicate['pred'] in digraph.nodes():
        # association_has_subject
        for neighbor, edges in digraph.pred[sub].items():
            for edge in edges.values():
                if edge['pred'] == 'OBAN:association_has_subject':
                    subject_assocs.add(neighbor)

        # association_has_object
        for neighbor, edges in digraph.pred[obj].items():
            for edge in edges.values():
                if edge['pred'] == 'OBAN:association_has_object':
                    object_assocs.add(neighbor)

        # association_has_predicate
        for neighbor, edges in digraph.pred[predicate['pred']].items():
            for edge in edges.values():
                if edge['pred'] == 'OBAN:association_has_predicate':
                    predicate_assocs.add(neighbor)

        union_assoc = subject_assocs & object_assocs & predicate_assocs

    return union_assoc


def _triple_to_association(digraph, subject, predicate, obj):
    """
    Convert triple to association object
    """
    object_eq = []
    subject_eq = []
    if 'equivalentOriginalNodeTarget' in predicate:
        for eq in predicate['equivalentOriginalNodeTarget']:
            curies = contract_uri(eq, [get_curie_map()], shortest=True)
            if len(curies) != 0:
                object_eq.append(curies[0])

    if 'equivalentOriginalNodeSource' in predicate:
        for eq in predicate['equivalentOriginalNodeSource']:
            curies = contract_uri(eq, [get_curie_map()], shortest=True)
            if len(curies) != 0:
                subject_eq.append(curies[0])

    relation_lbl = predicate['lbl'][0] if predicate['lbl'] else None

    association = {
        'subject': {
            'id': subject,
            'label': digraph.nodes[subject]['lbl'],
            'iri': expand_uri(subject, [get_curie_map()])
        },
        'subject_eq': subject_eq,
        'relation': {
            'id': predicate['pred'],
            'label': relation_lbl,
            'iri': expand_uri(predicate['pred'], [get_curie_map()])
        },
        'object': {
            'id': obj,
            'label': digraph.nodes[obj]['lbl'],
            'iri': expand_uri(obj, [get_curie_map()])
        },
        'object_eq': object_eq,
        'provided_by': predicate['isDefinedBy'],
        'evidence_types': [],
        'publications': []
    }

    # get association node linked to ECO codes and publications
    association_nodes = _get_association_nodes(digraph, subject, predicate, obj)

    if len(list(association_nodes)) > 1:
        # This can happen with clique merging, for now log it
        # and combine both in association results
        logging.debug("Ambiguous association for %s, %s, %s",
                      subject, predicate, obj)

    for association_node in list(association_nodes):
        for obj, edges in digraph.adj[association_node].items():
            eco_codes = [eco['id'] for eco in association['evidence_types']]
            pubs = [pub['id'] for pub in association['publications']]

            for edge in edges.values():
                if edge['pred'] == 'RO:0002558' and obj not in eco_codes:
                    association['evidence_types'].append({
                        'id': obj,
                        'label': digraph.nodes[obj]['lbl']
                    })
                elif edge['pred'] == 'dc:source' and obj not in pubs:
                    association['publications'].append({
                        'id': obj,
                        'label': digraph.nodes[obj]['lbl']
                    })

    return association


def obograph_to_assoc_results(digraph, assoc_type, is_publication=False):
    """
    Converts a multidigraph from convert_json_object
    to a list of association objects, which is easier
    to parse downstream for graph and table views

    :param: digraph: networkx.multidigraph or digraph
    :return: List, list of golr association objects

    """
    # Filter out oban edges initially, then retrieve using
    # _triple_to_association
    filter_edges = [
        'OBAN:association_has_subject',
        'OBAN:association_has_object',
        'OBAN:association_has_predicate',
        'RO:0002558'  # ECO codes, etc
    ]
    if not assoc_type.startswith('publication'):
        # publications, web pages, etc
        filter_edges.append('dc:source')

    # Filter out has_affected_feature since there can be many variant
    # to disease and the variant to gene assoc is usually obvious
    if assoc_type == 'gene_disease':
        filter_edges.append('GENO:0000418')

    association_results = []

    # Neighbors is a dict of adjacent nodes and outgoing edges
    for sub, neighbors in digraph.adjacency():
        # Iterate over nodes connected to outgoing edges (objects)
        for obj, edges_dict in neighbors.items():
            # Iterate over all edges connecting subject and object
            for edge_attr in edges_dict.values():
                if edge_attr['pred'] in filter_edges:
                    continue
                association_results.append(_triple_to_association(
                    digraph, sub, edge_attr, obj))

    return association_results


@cache.memoize()
def get_evidence_tables(id, is_publication, user_agent):

    results = search_associations(
            fq={'id': id},
            facet=False,
            select_fields=['evidence_graph', 'association_type'],
            user_agent=user_agent)
    assoc_results = {}
    assoc = results['associations'][0] if len(results['associations']) > 0 else {}
    if assoc:
        eg = {'graphs': [assoc.get('evidence_graph')]}
        assoc_type = assoc['type']
        digraph = convert_json_object(eg, reverse_edges=False)['graph']
        assoc_results = obograph_to_assoc_results(digraph, assoc_type, is_publication)
        assoc_results = {
            'associations': assoc_results,
            'numFound': len(assoc_results)
        }
    return assoc_results
